"""Provides various utility functions for BCML operations"""
# Copyright 2019 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
#import csv
import json
import os
import re
import shutil
import subprocess
import sys
import traceback
import unicodedata
import urllib.error
import urllib.request
from collections import namedtuple, OrderedDict
from collections.abc import Mapping
from configparser import ConfigParser
from pathlib import Path
from typing import Union
from PySide2.QtGui import QIcon, QPixmap

import byml
from byml import yaml_util
import sarc
import syaz0
import xxhash
import yaml

BcmlMod = namedtuple('BcmlMod', 'name priority path')
CREATE_NO_WINDOW = 0x08000000
SARC_EXTS = {'.sarc', '.pack', '.bactorpack', '.bmodelsh', '.beventpack', '.stera', '.stats',
             '.ssarc', '.spack', '.sbactorpack', '.sbmodelsh', '.sbeventpack', '.sstera', '.sstats'}
AAMP_EXTS = {'.bxml', '.sbxml', '.bas', '.sbas', '.baglblm', '.sbaglblm', '.baglccr', '.sbaglccr',
             '.baglclwd', '.sbaglclwd', '.baglcube', '.sbaglcube', '.bagldof', '.sbagldof',
             '.baglenv', '.sbaglenv', '.baglenvset', '.sbaglenvset', '.baglfila', '.sbaglfila',
             '.bagllmap', '.sbagllmap', '.bagllref', '.sbagllref', '.baglmf', '.sbaglmf',
             '.baglshpp', '.sbaglshpp', '.baiprog', '.sbaiprog', '.baslist', '.sbaslist',
             '.bassetting', '.sbassetting', '.batcl', '.sbatcl', '.batcllist', '.sbatcllist',
             '.bawareness', '.sbawareness', '.bawntable', '.sbawntable', '.bbonectrl',
             '.sbbonectrl', '.bchemical', '.sbchemical', '.bchmres', '.sbchmres', '.bdemo',
             '.sbdemo', '.bdgnenv', '.sbdgnenv', '.bdmgparam', '.sbdmgparam', '.bdrop', '.sbdrop',
             '.bgapkginfo', '.sbgapkginfo', '.bgapkglist', '.sbgapkglist', '.bgenv', '.sbgenv',
             '.bglght', '.sbglght', '.bgmsconf', '.sbgmsconf', '.bgparamlist', '.sbgparamlist',
             '.bgsdw', '.sbgsdw', '.bksky', '.sbksky', '.blifecondition', '.sblifecondition',
             '.blod', '.sblod', '.bmodellist', '.sbmodellist', '.bmscdef', '.sbmscdef', '.bmscinfo',
             '.sbmscinfo', '.bnetfp', '.sbnetfp', '.bphyscharcon', '.sbphyscharcon',
             '.bphyscontact', '.sbphyscontact', '.bphysics', '.sbphysics', '.bphyslayer',
             '.sbphyslayer', '.bphysmaterial', '.sbphysmaterial', '.bphyssb', '.sbphyssb',
             '.bphyssubmat', '.sbphyssubmat', '.bptclconf', '.sbptclconf', '.brecipe', '.sbrecipe',
             '.brgbw', '.sbrgbw', '.brgcon', '.sbrgcon', '.brgconfig', '.sbrgconfig',
             '.brgconfiglist', '.sbrgconfiglist', '.bsfbt', '.sbsfbt', '.bsft', '.sbsft', '.bshop',
             '.sbshop', '.bumii', '.sbumii', '.bvege', '.sbvege', '.bactcapt', '.sbactcapt'}
BYML_EXTS = {'.bgdata', '.sbgdata', '.bquestpack', '.sbquestpack', '.byml', '.sbyml', '.mubin',
             '.smubin', '.baischedule', '.sbaischedule', '.baniminfo', '.sbaniminfo', '.bgsvdata',
             '.sbgsvdata'}


decompress = syaz0.decompress
compress = syaz0.compress


def get_exec_dir() -> Path:
    """ Gets the root BCML directory """
    return Path(os.path.dirname(os.path.realpath(__file__)))


def get_data_dir() -> Path:
    """ Gets BCML's local app data directory """
    data_dir = Path(os.path.expandvars('%LOCALAPPDATA%')) / 'bcml'
    if not data_dir.exists():
        data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def get_work_dir() -> Path:
    """ Gets the BCML internal working directory """
    work_dir = get_data_dir() / 'work_dir'
    if not work_dir.exists():
        work_dir.mkdir(parents=True, exist_ok=True)
    return work_dir


def clear_temp_dir():
    """Empties BCML's temp directories"""
    for path in get_work_dir().glob('tmp*'):
        try:
            if path.is_dir():
                shutil.rmtree(str(path))
            elif path.is_file():
                path.unlink()
        except OSError:
            pass


def get_icon(name: str) -> QIcon:
    """ Gets the specified BCML tool icon """
    icon = QIcon()
    icon.addPixmap(QPixmap(str(get_exec_dir() / 'data' / f'{name}')))
    return icon


def get_settings() -> {}:
    """ Gets the BCML settings as a dict """
    if not hasattr(get_settings, 'settings_file'):
        settings = ConfigParser()
        settings_path = get_data_dir() / 'settings.ini'
        if not settings_path.exists():
            settings['Settings'] = {
                'cemu_dir': '',
                'game_dir': '',
                'load_reverse': False,
                'mlc_dir': '',
                'site_meta': '',
                'dark_theme': False,
                'guess_merge': False,
                'lang': ''
            }
            with settings_path.open('w', encoding='utf-8') as s_file:
                settings.write(s_file)
        else:
            settings.read(str(settings_path))
        get_settings.settings_file = settings
    return get_settings.settings_file['Settings']


def get_settings_bool(setting: str) -> bool:
    """Gets the value of a boolean setting"""
    if not setting in get_settings():
        return False
    return get_settings()[setting] == 'True'


def set_settings_bool(setting: str, value: bool):
    """Sets the value of a boolean setting"""
    get_settings.settings_file['Settings'][setting] = str(value)
    save_settings()


def save_settings():
    """Saves changes made to settings"""
    with (get_data_dir() / 'settings.ini').open('w', encoding='utf-8') as s_file:
        get_settings.settings_file.write(s_file)


def get_cemu_dir() -> Path:
    """ Gets the saved Cemu installation directory """
    cemu_dir = str(get_settings()['cemu_dir'])
    if not cemu_dir or not Path(cemu_dir).is_dir():
        err = FileNotFoundError('The Cemu directory has moved or not been saved yet.')
        err.error_text = 'The Cemu directory has moved or not been saved yet.'
        raise err
    return Path(cemu_dir)


def set_cemu_dir(path: Path):
    """ Sets the saved Cemu installation directory """
    settings = get_settings()
    settings['cemu_dir'] = str(path.resolve())
    save_settings()


def get_game_dir() -> Path:
    """ Gets the saved Breath of the Wild game directory """
    game_dir = str(get_settings()['game_dir'])
    if not game_dir or not Path(game_dir).is_dir():
        err = FileNotFoundError('The BotW game directory has has moved or not been saved yet.')
        err.error_text = 'The BotW game directory has has moved or not been saved yet.'
        raise err
    else:
        return Path(game_dir)


def set_game_dir(path: Path):
    """ Sets the saved Breath of the Wild game directory """
    settings = get_settings()
    settings['game_dir'] = str(path.resolve())
    save_settings()
    try:
        get_mlc_dir()
    except FileNotFoundError:
        try:
            from xml.dom import minidom
            set_path = get_cemu_dir() / 'settings.xml'
            if not set_path.exists():
                err = FileNotFoundError('The Cemu settings file could not be found.')
                err.error_text = 'The Cemu settings file could not be found. This usually means your Cemu directory '\
                                 'is set incorrectly.'
                raise err
            set_read = ''
            with set_path.open('r') as setfile:
                for line in setfile:
                    set_read += line.strip()
            settings = minidom.parseString(set_read)
            mlc_path = Path(settings.getElementsByTagName('mlc_path')[0].firstChild.nodeValue)
        except (FileNotFoundError, IndexError, ValueError, AttributeError):
            mlc_path = get_cemu_dir() / 'mlc01'
        if mlc_path.exists():
            set_mlc_dir(mlc_path)
        else:
            raise FileNotFoundError('The MLC directory could not be automatically located.')


def get_mlc_dir() -> Path:
    """ Gets the saved Cemu mlc directory """
    mlc_dir = str(get_settings()['mlc_dir'])
    if not mlc_dir or not Path(mlc_dir).is_dir():
        err = FileNotFoundError('The Cemu MLC directory has moved or not been saved yet.')
        err.error_text = 'The Cemu MLC directory has moved or not been saved yet.'
        raise err
    return Path(mlc_dir)


def set_mlc_dir(path: Path):
    """ Sets the saved Cemu mlc directory """
    settings = get_settings()
    settings['mlc_dir'] = str(path.resolve())
    save_settings()
    if hasattr(get_update_dir, 'update_dir'):
        del get_update_dir.update_dir
    if hasattr(get_aoc_dir, 'aoc_dir'):
        del get_aoc_dir.aoc_dir


def set_site_meta(site_meta):
    """ Caches site meta from url's specified in mods rules.txt """
    settings = get_settings()
    if not 'site_meta' in settings:
        settings['site_meta'] = ''
    else:
        settings['site_meta'] = str(settings['site_meta'] + f'{site_meta};')
    save_settings()


def get_title_id() -> (str, str):
    """Gets the title ID of the BotW game dump"""
    if not hasattr(get_title_id, 'title_id'):
        title_id = '00050000101C9400'
        with (get_game_dir().parent / 'code' / 'app.xml').open('r') as a_file:
            for line in a_file:
                title_match = re.search(
                    r'<title_id type=\"hexBinary\" length=\"8\">([0-9A-F]{16})</title_id>', line)
                if title_match:
                    title_id = title_match.group(1)
                    break
        get_title_id.title_id = (title_id[0:7] + '0', title_id[8:])
    return get_title_id.title_id


def get_update_dir() -> Path:
    """ Gets the path to the game's update files in the Cemu mlc directory """
    if not hasattr(get_update_dir, 'update_dir'):
        title_id = get_title_id()
        # First try the 1.15.11c mlc layout
        if (get_mlc_dir() / 'usr' / 'title' / f'{title_id[0][0:7]}E' / title_id[1] / 'content')\
            .exists():
            get_update_dir.update_dir = get_mlc_dir() / 'usr' / 'title' / \
                f'{title_id[0][0:7]}E' / title_id[1] / 'content'
        # Then try the legacy layout
        elif (get_mlc_dir() / 'usr' / 'title' / title_id[0] / title_id[1] / 'content').exists():
            get_update_dir.update_dir = get_mlc_dir() / 'usr' / 'title' / \
                title_id[0] / title_id[1] / 'content'
        else:
            err = FileNotFoundError('The Cemu update directory could not be found.')
            err.error_text = ('The Cemu update directory could not be found. Usually, the correct location is somewhere '
                              'in Cemu\'s MLC folder. This probably means that the MLC folder has not been correctly '
                              'located, or that you have not installed the update files in Cemu.')
            raise err
    return get_update_dir.update_dir


def get_aoc_dir() -> Path:
    """ Gets the path to the game's aoc files in the Cemu mlc direcroy """
    if not hasattr(get_aoc_dir, 'aoc_dir'):
        title_id = get_title_id()
        mlc_title = get_mlc_dir() / 'usr' / 'title'
        # First try the 1.15.11c mlc layout
        if (mlc_title / f'{title_id[0][0:7]}C' / title_id[1] / 'content' / '0010').exists():
            get_aoc_dir.aoc_dir = get_mlc_dir() / 'usr' / 'title' / \
                f'{title_id[0][0:7]}C' / title_id[1] / 'content' / '0010'
        # Then try the legacy layout
        elif (mlc_title / title_id[0] / title_id[1] / 'aoc' / 'content' / '0010').exists():
            get_aoc_dir.aoc_dir = get_mlc_dir() / 'usr' / 'title' / \
                title_id[0] / title_id[1] / 'aoc' / 'content' / '0010'
        else:
            err = FileNotFoundError('The Cemu aoc directory could not be found.')
            err.error_text = ('The Cemu aoc directory could not be found. Usually, the correct location is somewhere '
                              'in Cemu\'s MLC folder. This probably means that the MLC folder has not been correctly '
                              'located, or that you have not installed the DLC files in Cemu.')
            raise err
    return get_aoc_dir.aoc_dir


def get_modpack_dir() -> Path:
    """ Gets the Cemu graphic pack directory for mods """
    if 'Beta' not in get_bcml_version():
        return get_cemu_dir() / 'graphicPacks' / 'BCML'
    else:
        return get_cemu_dir() / 'graphicPacks' / 'BCML_beta'


def get_util_dirs() -> tuple:
    """
    Gets the primary directories BCML uses

    :returns: A tuple containing the root BCML directory, the BCML working
    directory, the Cemu installation directory, and the Cemu graphicPacks
    directory.
    :rtype: (class:`pathlib.Path`, class:`pathlib.Path`, class:`pathlib.Path`,
            class:`pathlib.Path`)
    """
    return get_exec_dir(), get_work_dir(), get_cemu_dir(), get_modpack_dir()


def get_botw_dirs() -> tuple:
    """
    Gets the directories the BotW game files

    :returns: A tuple containing the main BotW directory, the update directoy,
    and the aoc directory.
    :rtype: (class:`pathlib.Path`, class:`pathlib.Path`, class:`pathlib.Path`)
    """
    return get_game_dir(), get_update_dir(), get_aoc_dir()


def get_bcml_version() -> str:
    """Gets the version string for the installed copy of BCML"""
    from .__version__ import VERSION
    return VERSION + (' Beta' if 'b' in VERSION else '')


def get_game_file(path: Union[Path, str], aoc: bool = False) -> Path:
    """
    Gets the path to an original copy of a modded file from the game dump.

    :param path: The relative path to the modded file.
    :type path: Union[class:`pathlib.Path`, str]
    :param aoc: Whether the file is part of add-on content (DLC)
    :type aoc: bool, optional
    """
    if str(path).startswith('content/') or str(path).startswith('content\\'):
        path = Path(str(path).replace('content/', '').replace('content\\', ''))
    if isinstance(path, str):
        path = Path(path)
    game_dir = get_game_dir()
    update_dir = get_update_dir()
    try:
        aoc_dir = get_aoc_dir()
    except FileNotFoundError:
        aoc_dir = None
    if 'aoc' in str(path) or aoc:
        if aoc_dir:
            path = Path(
                path.as_posix().replace('aoc/content/0010/', '').replace('aoc/0010/content/', '')
                .replace('aoc/content/', '').replace('aoc/0010/', '')
            )
            if (aoc_dir / path).exists():
                return aoc_dir / path
            raise FileNotFoundError(f'{path} not found in DLC files.')
        else:
            raise FileNotFoundError(f'{path} is a DLC file, but the DLC directory is missing.')
    if (update_dir / path).exists():
        return update_dir / path
    if (game_dir / path).exists():
        return game_dir / path
    elif aoc_dir and (aoc_dir / path).exists():
        return aoc_dir / path
    else:
        raise FileNotFoundError(f'File {str(path)} was not found in game dump.')


def get_nested_file_bytes(file: str, unyaz: bool = True) -> bytes:
    """
    Get the contents of a file nested inside one or more SARCs

    :param file: A string containing the nested SARC path to the file
    :type file: str
    :param unyaz: Whether to decompress the file if yaz0 compressed, defaults to True
    :type unyaz: bool, optional
    :return: Returns the bytes to the file
    :rtype: bytes
    """
    nests = file.split('//')
    sarcs = []
    with open(nests[0], 'rb') as s_file:
        sarcs.append(sarc.read_file_and_make_sarc(s_file))
    i = 1
    while i < len(nests) - 1:
        sarc_bytes = unyaz_if_needed(
            sarcs[i - 1].get_file_data(nests[i]).tobytes())
        sarcs.append(sarc.SARC(sarc_bytes))
        i += 1
    file_bytes = sarcs[-1].get_file_data(nests[-1]).tobytes()
    if file_bytes[0:4] == b'Yaz0' and unyaz:
        file_bytes = decompress(file_bytes)
    del sarcs
    return file_bytes


def get_master_modpack_dir() -> Path:
    """ Gets the directory for the BCML master graphicpack """
    master = get_modpack_dir() / '9999_BCML'
    if not (master / 'rules.txt').exists():
        create_bcml_graphicpack_if_needed()
    return master


def get_hash_table() -> {}:
    """ Returns a dict containing an xxHash table for BotW game files """
    if not hasattr(get_hash_table, 'table'):
        with (get_exec_dir() / 'data' / 'hashtable.json').open('r') as h_file:
            get_hash_table.table = json.load(h_file)
    return get_hash_table.table


def get_canon_name(file: str, allow_no_source: bool = False) -> str:
    """ Gets the canonical path of a game file taken from an extracted graphic pack """
    name = str(file).replace("\\", "/").replace('.s', '.')\
        .replace('Content', 'content').replace('Aoc', 'aoc')
    if 'aoc/' in name:
        return name.replace('aoc/content', 'aoc').replace('aoc', 'Aoc')
    elif 'content/' in name and '/aoc' not in name:
        return name.replace('content/', '')
    elif allow_no_source:
        return name


def get_mod_id(mod_name: str, priority: int) -> str:
    """ Gets the ID for a mod from its name and priority """
    return f'{priority:04}_' + re.sub(r'(?u)[^-\w.]', '', mod_name.strip().replace(' ', ''))


def get_mod_by_priority(priority: int) -> Union[Path, bool]:
    """ Gets the path to the modpack installed with a given priority, or False if there is none """
    try:
        return list(get_modpack_dir().glob(f'{priority:04}*'))[0]
    except IndexError:
        return False


def is_pack_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects pack merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'packs.log').exists()


def is_gamedata_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects game data merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'gamedata.yml').exists()


def is_savedata_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects save data merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'savedata.yml').exists()


def is_actorinfo_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects actor info merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'actorinfo.yml').exists()


def is_eventinfo_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects event info merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'eventinfo.yml').exists()


def is_map_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects map merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'map.yml').exists()


def is_deepmerge_mod(mod: Union[Path, BcmlMod, str]) -> bool:
    """ Checks whether a mod affects deep merging """
    path = mod.path if isinstance(mod, BcmlMod) else Path(
        mod) if isinstance(mod, str) else mod
    return (path / 'logs' / 'deepmerge.yml').exists()


def get_file_language(file: Union[Path, str]) -> str:
    """ Extracts the game language of a file from its name """
    if isinstance(file, Path):
        file = str(file)
    lang_match = re.search(r'_([A-Z]{2}[a-z]{2})', file)
    return lang_match.group(1)


def is_file_modded(name: str, file: Union[bytes, Path], count_new: bool = True) -> bool:
    """
    Determines if a game file has been modified by checking the hash

    :param name: Canonical path to the file being checked.
    :type name: str
    :param file: Bytes representing the file contents.
    :type file: Union[bytes, Path]
    :param count_new: Whether to count new files as modded, defaults to False
    :type count_new: bool, optional
    :returns: True if the file hash differs from the entry in the table or
    if the entry does not exist, false if it matches the table entry.
    :rtype: bool
    """
    contents = file if isinstance(file, bytes) else \
        file.read_bytes() if isinstance(file, Path) else file.tobytes()
    table = get_hash_table()
    if name not in table:
        return count_new
    fhash = xxhash.xxh32(contents).hexdigest()
    return not fhash in table[name]


def is_yaml_modded(entry, ref_list: dict, mod_list: dict) -> bool:
    """
    Determines if a YAML object has been modified from the original

    :param entry: The key of the YAML entry to compare.
    :type entry: str
    :param ref_list: The reference YAML dictionary.
    :type ref_list: dict
    :param mod_list: The modded YAML dictionary.
    :type mod_list: dict
    :returns: True if the normalized entry contents differs from the entry
    in the reference document or if the entry does not exist, false if it
    matches the reference entry.
    :rtype: bool
    """
    mod_entry = unicodedata.normalize('NFC', mod_list['entries'][entry].__str__())
    mod_entry = re.sub('[^0-9a-zA-Z]+', '', mod_entry)
    try:
        ref_entry = unicodedata.normalize('NFC', ref_list['entries'][entry].__str__())
        ref_entry = re.sub('[^0-9a-zA-Z]+', '', ref_entry)
    except KeyError:
        return True
    if not ref_entry == mod_entry:
        return True
    return False


def byml_to_yml_dir(tmp_dir: Path, ext: str = '.byml'):
    """ Converts BYML files in given temp dir to YAML """
    dumper = yaml.CDumper
    yaml_util.add_representers(dumper)
    for data in tmp_dir.rglob(f'**/*{ext}'):
        yml_data = byml.Byml(data.read_bytes())
        with (data.with_name(data.stem + '.yml')).open('w', encoding='utf-8') as y_file:
            yaml.dump(yml_data.parse(), y_file, Dumper=dumper,
                      allow_unicode=True, encoding='utf-8')
        data.unlink()


def yml_to_byml_dir(tmp_dir: Path, ext: str = '.byml'):
    """ Converts YAML files in given temp dir to BYML """
    loader = yaml.CSafeLoader
    yaml_util.add_constructors(loader)
    for yml in tmp_dir.rglob('**/*.yml'):
        with yml.open('r', encoding='utf-8') as y_file:
            root = yaml.load(y_file, loader)
        with (yml.with_name(yml.stem + ext)).open('wb') as b_file:
            byml.Writer(root, True).write(b_file)
        yml.unlink()


def is_file_sarc(path: str) -> bool:
    """ Checks the file extension of a game file to tell if it is a SARC """
    ext = os.path.splitext(str(path))[1]
    return ext in SARC_EXTS


def is_file_aamp(path: str) -> bool:
    """ Checks the file extension of a game file to tell if it is an AAMP file """
    ext = os.path.splitext(str(path))[1]
    return ext in AAMP_EXTS


def is_file_byml(path: str) -> bool:
    """ Checks the file extension of a game file to tell if it is a BYML file """
    ext = os.path.splitext(str(path))[1]
    return ext in BYML_EXTS


def decompress_file(file) -> bytes:
    if isinstance(file, str):
        file = Path(file)
    return decompress(file.read_bytes())


def unyaz_if_needed(file_bytes: bytes) -> bytes:
    """
    Detects by file extension if a file should be decompressed, and decompresses if needed

    :param file_bytes: The bytes to potentially decompress.
    :type file_bytes: bytes
    :returns: Returns the bytes of the file, decompressed if necessary.
    :rtype: bytes
    """
    if file_bytes[0:4] == b'Yaz0':
        return decompress(file_bytes)
    else:
        return file_bytes


def inject_file_into_bootup(file: str, data: bytes, create_bootup: bool = False):
    """
    Injects a file into the master BCML `Bootup.pack`

    :param file: The path of the file to inject
    :type file: str
    :param data: The bytes of the file to inject
    :type data: bytes
    :param create_bootup: Whether to create `Bootup.pack` if it does not exist, defaults to False
    :type create_bootup: bool, optional
    """
    bootup_path = get_master_modpack_dir() / 'content' / 'Pack' / 'Bootup.pack'
    if bootup_path.exists() or create_bootup:
        if not bootup_path.exists():
            bootup_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(get_game_file('Pack/Bootup.pack'), bootup_path)
        with bootup_path.open('rb') as b_file:
            old_bootup = sarc.read_file_and_make_sarc(b_file)
            new_bootup = sarc.make_writer_from_sarc(old_bootup)
        if file in old_bootup.list_files():
            new_bootup.delete_file(file)
        new_bootup.add_file(file, data)
        bootup_path.write_bytes(new_bootup.get_bytes())
    else:
        raise FileNotFoundError('Bootup.pack is not present in the master BCML mod')


def get_mod_info(rules_path: Path) -> BcmlMod:
    """ Gets the name and priority of a mod from its rules.txt """
    rules: ConfigParser = RulesParser()
    rules.read(str(rules_path))
    name = str(rules['Definition']['name']).strip('" \'').replace('_', ' ')
    try:
        priority = int(rules['Definition']['fsPriority'])
    except KeyError:
        error = KeyError(
            f'Mod "{name}" contains no priority field. It may not have been installed completely.'
        )
        error.path = rules_path.parent
        raise error
    return BcmlMod(
        name,
        priority,
        rules_path.parent
    )


def get_mod_preview(mod: BcmlMod, rules: ConfigParser = None) -> QPixmap:
    """
    Gets the preview image of a given mod, if any, and caches it

    :param mod: The mod to preview
    :type mod: class:`bcml.util.BcmlMod`
    :param rules: The contents of the mod's `rules.txt` file
    :type rules: ConfigParser
    :return: Returns the preview image for the mod as QPixmap
    :rtype: QPixmap
    """
    if not rules:
        rules = RulesParser()
        rules.read(str(mod.path / 'rules.txt'))
    if 'url' in rules['Definition']:
        url = str(rules['Definition']['url'])
    if not list(mod.path.glob('thumbnail.*')):
        if 'image' not in rules['Definition']:
            if 'url' in rules['Definition'] and 'gamebanana.com' in url:
                response = urllib.request.urlopen(url)
                rdata = response.read().decode()
                img_match = re.search(
                    r'<meta property=\"og:image\" ?content=\"(.+?)\" />', rdata)
                if img_match:
                    image_path = 'thumbnail.jfif'
                    urllib.request.urlretrieve(
                        img_match.group(1),
                        str(mod.path / image_path)
                    )
                else:
                    raise IndexError(f'Rule for {url} failed to find the remote preview')
            else:
                raise KeyError(f'No preview image available')
        else:
            image_path = str(rules['Definition']['image'])
            if image_path.startswith('http'):
                urllib.request.urlretrieve(
                    image_path,
                    str(mod.path / ('thumbnail.' + image_path.split(".")[-1]))
                )
                image_path = 'thumbnail.' + image_path.split(".")[-1]
            if not os.path.isfile(str(mod.path / image_path)):
                raise FileNotFoundError(
                    f'Preview {image_path} specified in rules.txt not found')
    else:
        for thumb in mod.path.glob('thumbnail.*'):
            image_path = thumb
    return QPixmap(str(mod.path / image_path))


def get_mod_link_meta(rules: ConfigParser = None):
    """
    Gets the link metadata for the URL field of a mod

    :param rules: The contents of the mod's `rules.txt` file
    :type rules: ConfigParser
    :return: Returns an HTML link to the mod webpage
    :rtype: str
    """
    url = str(rules['Definition']['url'])
    mod_domain = ''
    if 'www.' in url:
        mod_domain = url.split('.')[1]
    elif 'http' in url:
        mod_domain = url.split('//')[1].split('.')[0]
    site_name = mod_domain.capitalize()
    fetch_site_meta = True
    if 'site_meta' not in get_settings():
        set_site_meta('')
    if len(get_settings()['site_meta'].split(';')) > 1:
        for site_meta in get_settings()['site_meta'].split(';'):
            if site_meta.split(':')[0] == mod_domain:
                fetch_site_meta = False
                site_name = site_meta.split(':')[1]
    if fetch_site_meta:
        try:
            response = urllib.request.urlopen(url)
            rdata = response.read().decode()
            name_match = re.search(
                r'property=\"og\:site_name\"[^\/\>]'
                r'*content\=\"(.+?)\"|content\=\"(.+?)\"[^\/\>]'
                r'*property=\"og\:site_name\"',
                rdata
            )
            if name_match:
                for group in name_match.groups():
                    if group is not None:
                        set_site_meta(f'{mod_domain}:{group}')
                        site_name = str(group)
            img_match = re.search(
                r'<link.*rel=\"(shortcut icon|icon)\".*href=\"(.+?)\".*>', rdata)
            if img_match:
                (get_exec_dir() / 'work_dir' / 'cache' / 'site_meta').mkdir(
                    parents=True,
                    exist_ok=True
                )
                try:
                    urllib.request.urlretrieve(
                        img_match.group(2),
                        str(get_exec_dir() / 'work_dir' / 'cache' / "site_meta" /\
                            f'fav_{site_name}.{img_match.group(2).split(".")[-1]}')
                    )
                except (urllib.error.URLError,
                        urllib.error.HTTPError,
                        urllib.error.ContentTooShortError):
                    pass
        except (urllib.error.URLError,
                urllib.error.HTTPError,
                urllib.error.ContentTooShortError):
            pass
    favicon = ''
    for file in (get_exec_dir() / "work_dir" / "cache" / "site_meta")\
                .glob(f'fav_{site_name}.*'):
        favicon = f'<img src="{file.resolve()}" height="16"/> '
    return f'<b>Link: <a style="text-decoration: none;" href="{url}">{favicon} {site_name}</a></b>'


def get_installed_mods(disabled: bool = False) -> []:
    """
    Gets all installed mods and their basic info

    :param disabled: Whether to include disabled mods, defaults to False
    :type disabled: bool, optional
    :returns: A list of mods with their names, priorities, and installed paths.
    :rtype: list of (str, int, class:`pathlib.Path`) )
    """
    mods = []
    for rules in get_modpack_dir().glob('*/rules.txt*'):
        if rules.parent.stem == '9999_BCML' or (not disabled and rules.suffix == '.disable'):
            continue
        try:
            mod = get_mod_info(rules)
        except KeyError as priority_error:
            rules.unlink()
            shutil.rmtree(str(priority_error.path), ignore_errors=True)
            continue
        mods.insert(mod.priority - 100, mod)
    return mods


def is_mod_disabled(mod: BcmlMod) -> bool:
    return (mod.path / 'rules.txt.disable').exists()


def get_all_modded_files(only_loose: bool = False) -> dict:
    """
    Gets all installed file modifications and the highest priority of each

    :return: A dict of canonical paths and the priority of the highest modded version
    :rtype: dict of str: int
    """
    modded_files = {}
    for mod in get_installed_mods():
        with (mod.path / 'logs' / 'rstb.log').open('r') as l_file:
            csv_loop = csv.reader(l_file)
            for row in csv_loop:
                if row[0] == 'name' or (only_loose and '//' in row[2]):
                    continue
                modded_files[row[0]] = mod.priority
    return modded_files


def log_error():
    """ Writes the most recent error traceback to the error log and prints the traceback text """
    log_path = get_work_dir() / 'error.log'
    error_log = traceback.format_exc()
    with log_path.open('w', encoding='utf-8') as l_file:
        l_file.write(error_log)
    print('BCML has encountered an error. The details are as follows:')
    print(error_log)
    print(f'The error information has been saved to:\n  {str(log_path)}')
    if sys.stdin.isatty():
        sys.exit(1)


def update_bcml():
    """ Updates BCML to the latest version """
    subprocess.call([sys.executable, '-m', 'pip', 'install', '--upgrade', 'bcml'])


def create_bcml_graphicpack_if_needed():
    """Creates the BCML master modpack if it doesn't exist"""
    bcml_mod_dir = get_modpack_dir() / '9999_BCML'
    (bcml_mod_dir / 'logs').mkdir(parents=True, exist_ok=True)
    rules = bcml_mod_dir / 'rules.txt'
    if not rules.exists():
        with rules.open('w', encoding='utf-8') as r_file:
            r_file.write('[Definition]\n'
                         'titleIds = 00050000101C9300,00050000101C9400,00050000101C9500\n'
                         'name = BCML\n'
                         'path = {BCML: DON\'T TOUCH}/Master BCML\n'
                         'description = Auto-generated pack which merges RSTB changes and packs '
                         'for other mods\n'
                         'version = 4\n'
                         'fsPriority = 9999')


def dict_merge(dct: dict, merge_dct: dict, overwrite_lists: bool = False):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :param overwrite_lists: Whether to prevent duplicate items in lists, defaults to False
    :return: None
    """
    for k in merge_dct:
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)):
            dict_merge(dct[k], merge_dct[k])
        elif (k in dct and isinstance(dct[k], list)
              and isinstance(merge_dct[k], list)):
            if overwrite_lists:
                dct[k] = merge_dct[k]
            else:
                dct[k].extend(merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def create_schema_handler():
    import winreg
    with winreg.CreateKey(winreg.HKEY_CURRENT_USER, r'Software\Classes\bcml') as key:
        try:
            winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Software\Classes\bcml\shell\open\command',
                           0, winreg.KEY_READ)
        except (WindowsError, OSError):
            winreg.SetValueEx(key, 'URL Protocol', 0, winreg.REG_SZ, '')
            with winreg.CreateKey(key, r'shell\open\command') as key2:
                if (Path(os.__file__).parent.parent / 'Scripts' / 'bcml.exe').exists():
                    exec_path = Path(os.__file__).parent.parent / 'Scripts' / 'bcml.exe'
                elif (Path(__file__).parent.parent.parent / 'bin' / 'bcml.exe'):
                    exec_path = (Path(__file__).parent.parent.parent / 'bin' / 'bcml.exe')
                else:
                    return
                winreg.SetValueEx(key2, '', 0, winreg.REG_SZ, f'"{exec_path.resolve()}" "%1"')


class RulesParser(ConfigParser):
    def __init__(self):
        ConfigParser.__init__(self, dict_type=MultiDict)

    def write(self, fileobject):
        from io import StringIO
        buf = StringIO()
        ConfigParser.write(self, buf)
        config_str = re.sub(r'\[Preset[0-9]+\]', '[Preset]', buf.getvalue())
        fileobject.write(config_str)


class MultiDict(OrderedDict):
    _unique = 0

    def __setitem__(self, key, val):
        if isinstance(val, dict) and key == 'Preset':
            self._unique += 1
            key += str(self._unique)
        OrderedDict.__setitem__(self, key, val)


"""Main BCML application"""
# Copyright 2019 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
# pylint: disable=invalid-name,missing-docstring,too-many-lines
import os
import platform
import shutil
import subprocess
import sys
import threading
import traceback
import urllib.error
import urllib.request
from configparser import ConfigParser
from pathlib import Path

from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtCore import QUrl
from PySide2.QtGui import QDesktopServices
from PySide2.QtWidgets import QFileDialog

from bcml import data, install, merge, pack, rstable, texts, util, mubin, events, mergers
from bcml.Ui_about import Ui_AboutDialog
from bcml.Ui_install import Ui_InstallDialog
from bcml.Ui_main import Ui_MainWindow
from bcml.Ui_progress import Ui_ProgressDialog
from bcml.Ui_settings import Ui_SettingsDialog
from bcml.Ui_package import Ui_PackageDialog
from bcml.Ui_options import Ui_OptionsDialog
from bcml.util import BcmlMod


class MainWindow(QtWidgets.QMainWindow, Ui_MainWindow):

    def __init__(self):
        super(MainWindow, self).__init__()
        self.setupUi(self)
        self._mods = []
        self._mod_infos = {}
        self._progress = None
        self._thread = None
        self._cemu_exe = None

        self.setWindowIcon(util.get_icon('bcml.ico'))
        load_reverse = util.get_settings_bool('load_reverse')

        self.menRemerge = QtWidgets.QMenu()
        self.actMergeRstb = QtWidgets.QAction('Remerge RSTB')
        self.actMergePacks = QtWidgets.QAction('Remerge Packs')
        self.actMergeTexts = QtWidgets.QAction('Remerge Texts')
        self.actMergeActors = QtWidgets.QAction('Remerge Actor Info')
        self.actMergeGamedata = QtWidgets.QAction('Remerge Game/Save Data')
        self.actMergeMaps = QtWidgets.QAction('Remerge Maps')
        self.actDeepMerge = QtWidgets.QAction('Refresh Deep Merge')
        self.actMergeRstb.triggered.connect(self.MergeRstb_Clicked)
        self.actMergePacks.triggered.connect(self.MergePacks_Clicked)
        self.actMergeTexts.triggered.connect(self.MergeTexts_Clicked)
        self.actMergeActors.triggered.connect(self.MergeActors_Clicked)
        self.actMergeGamedata.triggered.connect(self.MergeGamedata_Clicked)
        self.actMergeMaps.triggered.connect(self.MergeMaps_Clicked)
        self.actDeepMerge.triggered.connect(self.DeepMerge_Clicked)
        self.menRemerge.addAction(self.actMergeRstb)
        self.menRemerge.addAction(self.actMergePacks)
        self.menRemerge.addAction(self.actMergeTexts)
        self.menRemerge.addAction(self.actMergeActors)
        self.menRemerge.addAction(self.actMergeGamedata)
        self.menRemerge.addAction(self.actMergeMaps)
        self.menRemerge.addAction(self.actDeepMerge)
        self.btnRemerge.setMenu(self.menRemerge)

        self.btnPackage = QtWidgets.QToolButton(self.statusBar())
        self.btnPackage.setIcon(util.get_icon('pack.png'))
        self.btnPackage.setToolTip('Create BNP Mod')
        self.btnSettings = QtWidgets.QToolButton(self.statusBar())
        self.btnSettings.setIcon(util.get_icon('settings.png'))
        self.btnSettings.setToolTip('Settings')
        self.btnAbout = QtWidgets.QToolButton(self.statusBar())
        self.btnAbout.setIcon(util.get_icon('about.png'))
        self.btnAbout.setToolTip('About')

        self.statusBar().addPermanentWidget(
            QtWidgets.QLabel('Version ' + util.get_bcml_version()))
        self.statusBar().addPermanentWidget(self.btnPackage)
        self.statusBar().addPermanentWidget(self.btnSettings)
        self.statusBar().addPermanentWidget(self.btnAbout)

        logo_path = str(util.get_exec_dir() / 'data' / 'logo-smaller.png')
        self._logo = QtGui.QPixmap(logo_path)
        self.lblImage.setFixedSize(256, 144)
        self._order_icons = {
            False: util.get_icon('up.png'),
            True: util.get_icon('down.png')
        }
        self.btnOrder.setIcon(self._order_icons[load_reverse])

        higher = 'top' if not load_reverse else 'bottom'
        lower = 'bottom' if not load_reverse else 'top'
        self.listWidget.setToolTip(f'Drag and drop to change mod load order. Mods at the {higher}'
                                   f' of the list override mods at the {lower}.')

        # Bind events
        self.listWidget.setDragDropMode(
            QtWidgets.QAbstractItemView.InternalMove)
        self.listWidget.itemSelectionChanged.connect(self.SelectItem)
        self.listWidget.installEventFilter(self)
        self.btnOrder.clicked.connect(self.OrderClicked)
        self.btnInstall.clicked.connect(self.InstallClicked)
        self.btnRemerge.clicked.connect(self.RemergeClicked)
        self.btnChange.clicked.connect(self.ChangeClicked)
        self.btnExport.clicked.connect(self.ExportClicked)
        self.btnBackup.clicked.connect(self.BackupClicked)
        self.btnRestore.clicked.connect(self.RestoreClicked)
        self.btnCemu.clicked.connect(self.CemuClicked)
        self.btnRemoveAll.clicked.connect(self.RemoveAllClicked)
        self.btnUninstall.clicked.connect(self.UninstallClicked)
        self.btnDisableMod.clicked.connect(self.DisableModClicked)
        self.btnEnableMod.clicked.connect(self.EnableModClicked)
        self.btnExplore.clicked.connect(self.ExploreClicked)
        self.btnPackage.clicked.connect(self.PackageClicked)
        self.btnSettings.clicked.connect(self.SettingsClicked)
        self.btnAbout.clicked.connect(self.AboutClicked)

    def eventFilter(self, watched, event):
        if event.type() == QtCore.QEvent.ChildRemoved:
            mods = [self.listWidget.item(i).data(Qt.UserRole)
                    for i in range(self.listWidget.count())]
            if mods != self._mods:
                self.btnChange.setEnabled(True)
            else:
                self.btnChange.setEnabled(False)
        return super(MainWindow, self).eventFilter(watched, event)

    def SetupChecks(self):
        try:
            util.get_cemu_dir()
        except FileNotFoundError:
            QtWidgets.QMessageBox.information(
                self, 'First Time', 'It looks like this may be your first time running BCML.'
                                    'Please select the directory where Cemu is installed. You can '
                                    'tell it\'s right if it contains <code>Cemu.exe</code>.')
            folder = QFileDialog.getExistingDirectory(
                self, 'Select Cemu Directory')
            if folder:
                util.set_cemu_dir(Path(folder))
            else:
                sys.exit(0)
        try:
            util.get_game_dir()
        except FileNotFoundError:
            QtWidgets.QMessageBox.information(
                self, 'Dump Location', 'BCML needs to know the location of your game dump. '
                'Please select the "content" directory in your dumped copy of Breath of the Wild. '
                'Note this is <em>not</em> usually inside Cemu\'s MLC folder. This is where the '
                'base game in installed, not update or DLC data.')
            folder = '/'
            while not (Path(folder).parent / 'code' / 'app.xml').exists():
                folder = QFileDialog.getExistingDirectory(self, 'Select Game Dump Directory')
                if folder:
                    folder = Path(folder)
                    if (folder / 'content').exists():
                        folder = folder / 'content'
                    if not (folder.parent / 'code' / 'app.xml').exists():
                        QtWidgets.QMessageBox.warning(
                            self,
                            'Error',
                            'The chosen directory does not appear to be a valid dump. Please select'
                            'the "content" directory in your dumped copy of Breath of the Wild.'
                        )
                    else:
                        try:
                            util.set_game_dir(folder)
                        except FileNotFoundError:
                            QtWidgets.QMessageBox.information(
                                self,
                                'MLC Location',
                                'BCML could not detect the location of Cemu\'s MLC directory for '
                                'your game. You will need to specify it manually.'
                            )
                            mlc_folder = QFileDialog.getExistingDirectory(
                                self,
                                'Select Cemu MLC Directory'
                            )
                            if mlc_folder:
                                util.set_mlc_dir(Path(mlc_folder))
                                util.set_game_dir(Path(folder))                    
                                util.get_update_dir()
                                try:
                                    util.get_aoc_dir()
                                except FileNotFoundError:
                                    pass
                            else:
                                sys.exit(0)
                else:
                    sys.exit(0)
        try:
            util.get_update_dir()
        except FileNotFoundError:
            QtWidgets.QMessageBox.warning(
                self, 'Dump Location', 'BCML could not locate your game\'s update data. Usually,  '
                'this means that either the game dump folder or the MLC folder is not correctly set'
                '. Please go to your BCML settings and confirm the correct location of both '
                'folders. You will not be able to install any mods until this is corrected')
        if 'lang' not in util.get_settings() or not util.get_settings()['lang']:
            lang, okay = QtWidgets.QInputDialog.getItem(
                self,
                'Select Language',
                'Select the regional language you\nuse to play Breath of the Wild',
                texts.LANGUAGES,
                0,
                False,
                flags=QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowTitleHint
            )
            if okay and lang:
                util.get_settings()['lang'] = lang
                util.save_settings()
        self.LoadMods()

    def LoadMods(self):
        self.statusBar().showMessage('Loading mods...')
        self.listWidget.clear()
        self._mods = sorted(util.get_installed_mods(disabled=True), key=lambda imod: imod.priority,
                            reverse=not util.get_settings_bool('load_reverse'))
        for mod in self._mods:
            mod_item = QtWidgets.QListWidgetItem()
            mod_item.setText(mod.name)
            mod_item.setData(QtCore.Qt.UserRole, mod)
            if util.is_mod_disabled(mod):
                mod_item.setTextColor(QtGui.QColor(211, 47, 47))
            self.listWidget.addItem(mod_item)
        self._mod_infos = {}
        self.lblModInfo.linkActivated.connect(self.link)
        self.lblModInfo.setText('No mod selected')
        self.lblImage.setPixmap(self._logo)
        self.lblImage.setFixedSize(256, 144)
        self.statusBar().showMessage(f'{len(self._mods)} mods installed')
        if self._mods:
            self.btnRemerge.setEnabled(True)
            self.btnExport.setEnabled(True)
            self.btnBackup.setEnabled(True)
            self.btnRemoveAll.setEnabled(True)
        else:
            self.btnRemerge.setEnabled(False)
            self.btnExport.setEnabled(False)
            self.btnBackup.setEnabled(False)
            self.btnRemoveAll.setEnabled(False)

    def link(self, linkStr):
        QDesktopServices.openUrl(QUrl(linkStr))

    def SelectItem(self):
        if not self.listWidget.selectedItems():
            return
        mod = self.listWidget.selectedItems()[0].data(QtCore.Qt.UserRole)

        if mod not in self._mod_infos:
            rules = util.RulesParser()
            if util.is_mod_disabled(mod):
                rules.read(str(mod.path / 'rules.txt.disable'))
            else:
                rules.read(str(mod.path / 'rules.txt'))
            font_met = QtGui.QFontMetrics(self.lblModInfo.font())
            path = str(mod.path)
            while font_met.boundingRect(f'Path: {path}.....').width() >= self.lblModInfo.width():
                path = path[:-1]
            changes = []
            for merger in {cls() for cls in mergers.get_mergers()}:
                if merger.is_mod_logged(mod):
                    changes.append(merger.NAME if not merger.NAME == 'rstb' else 'RSTB')
            mod_info = [
                f'<b>Name</b>: {mod.name}',
                f'<b>Priority:</b> {mod.priority}',
                f'<b>Path:</b> {path}...',
                '<b>Description:</b> {}'.format(
                    str(rules["Definition"]["description"]).strip(' "\'')),
                f'<b>Changes:</b> {", ".join(changes)}'
            ]
            if util.is_mod_disabled(mod):
                mod_info.insert(0, '<b><span style="color:#d32f2f">DISABLED</span></b>')
            if 'url' in rules['Definition']:
                mod_info.insert(3, util.get_mod_link_meta(rules))
            try:
                preview = util.get_mod_preview(mod, rules)
            except (FileNotFoundError, KeyError, IndexError, UnboundLocalError):
                preview = self._logo
            finally:
                self._mod_infos[mod] = (mod_info, preview)
        else:
            mod_info, preview = self._mod_infos[mod]

        self.lblModInfo.setText(f'<p>{"</p><p>".join(mod_info)}</p>')
        self.lblImage.setFixedSize(
            256, 256 // (preview.width() / preview.height())
        )
        self.lblImage.setPixmap(preview)
        self.lblModInfo.setWordWrap(True)

        if not self.btnUninstall.isEnabled():
            self.btnUninstall.setEnabled(True)
        if not self.btnExplore.isEnabled():
            self.btnExplore.setEnabled(True)
        disabled = util.is_mod_disabled(mod)
        self.btnEnableMod.setVisible(disabled)
        self.btnDisableMod.setVisible(not disabled)
        if not self.btnEnableMod.isEnabled():
            self.btnEnableMod.setEnabled(True)
        if not self.btnDisableMod.isEnabled():
            self.btnDisableMod.setEnabled(True)

    def PerformOperation(self, func, *args, title: str = 'Operation In Progress'):
        self.btnInstall.setEnabled(False)
        self.btnRemerge.setEnabled(False)
        self.btnExport.setEnabled(False)
        self.btnUninstall.setEnabled(False)
        self.btnEnableMod.setEnabled(False)
        self.btnDisableMod.setEnabled(False)
        self.btnExplore.setEnabled(False)
        self.btnBackup.setEnabled(False)
        self.btnRestore.setEnabled(False)
        self.btnCemu.setEnabled(False)
        self.btnRemoveAll.setEnabled(False)
        self.btnOrder.setEnabled(False)
        self._progress = ProgressDialog(self)
        self._progress.setWindowTitle(title)
        self._progress.setWindowModality(QtCore.Qt.ApplicationModal)
        self._progress.show()
        self._thread = ProgressThread(self._progress.lblProgress, func, *args)
        self._thread.start()
        self._thread.signal.sig.connect(self.OperationFinished)
        self._thread.signal.err.connect(self.OperationError)


    def OperationFinished(self):
        title = self._progress.windowTitle()
        self._progress.close()
        del self._progress
        self.btnInstall.setEnabled(True)
        self.btnRemerge.setEnabled(True)
        self.btnExport.setEnabled(True)
        self.btnRestore.setEnabled(True)
        self.btnCemu.setEnabled(True)
        self.btnOrder.setEnabled(True)
        self.LoadMods()
        msg_done = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Information,
            f'{title} Finished',
            'Operation complete!',
            parent=self
        )
        msg_done.addButton('OK', QtWidgets.QMessageBox.NoRole)
        msg_done.addButton('Launch Game', QtWidgets.QMessageBox.ActionRole)
        result = msg_done.exec_()
        del self._thread
        util.clear_temp_dir()
        if result == 1:
            self.CemuClicked()

    def OperationError(self):
        self.btnInstall.setEnabled(True)
        self.btnRemerge.setEnabled(True)
        self.btnExport.setEnabled(True)
        self.btnRestore.setEnabled(True)
        self.btnCemu.setEnabled(True)
        self.btnOrder.setEnabled(True)
        QtWidgets.QMessageBox.critical(
            self,
            'Error',
            f'BCML has encountered an error while performing an operation. '
            f'Error details:\n\n{self._thread.error}'
        )
        self.LoadMods()
        self._progress.close()
        del self._progress
        del self._thread
        util.clear_temp_dir()

    # Button handlers

    def OrderClicked(self):
        load_reverse = not util.get_settings_bool('load_reverse')
        util.set_settings_bool('load_reverse', load_reverse)
        self.LoadMods()
        self.btnOrder.setIcon(self._order_icons[load_reverse])

        higher = 'top' if not load_reverse else 'bottom'
        lower = 'bottom' if not load_reverse else 'top'
        self.listWidget.setToolTip('Drag and drop to change mod load order. Mods at the'
                                   f'{higher} of the list override mods at the {lower}.')

    def InstallClicked(self, preload_mod: Path = None):
        def install_all(result: dict):
            mods = []
            for i, mod in enumerate(result['paths']):
                mods.append(install.install_mod(
                    mod,
                    verbose=False,
                    options=result['options'],
                    wait_merge=True,
                    insert_priority=result['insert_priority'] + i
                ))
            all_mergers = [merge_class() for merge_class in mergers.get_mergers()]
            merges = set()
            partials = {}
            for mod in mods:
                for merger in all_mergers:
                    if merger.is_mod_logged(mod):
                        merges.add(merger)
                        if merger.can_partial_remerge():
                            if not merger.NAME in partials:
                                partials[merger.NAME] = set()
                            partials[merger.NAME] |= set(merger.get_mod_affected(mod))
            for merger in mergers.sort_mergers(merges):
                if merger.NAME in result['options']:
                    merger.set_options(result['options'][merger.NAME])
                if merger.NAME in partials:
                    merger.set_options({'only_these': partials[merger.NAME]})
                merger.perform_merge()

        dialog = InstallDialog(self)
        if preload_mod:
            mod_item = QtWidgets.QListWidgetItem()
            mod_item.setText(preload_mod.stem)
            mod_item.setData(Qt.UserRole, preload_mod)
            dialog.lstQueue.addItem(mod_item)
        result = dialog.GetResult()
        if result:
            self.PerformOperation(install_all, result, title='Installing')

    def RemergeClicked(self):
        self.PerformOperation(install.refresh_merges, ())


    def MergeRstb_Clicked(self):
        def full_rstb():
            rstable.log_merged_files_rstb()
            rstable.generate_master_rstb()
        self.PerformOperation(full_rstb, title='Remerging RSTB')


    def MergePacks_Clicked(self):
        self.PerformOperation(
            pack.merge_installed_packs,
            (False, None, False, True),
            title='Remerging Packs'
        )


    def MergeTexts_Clicked(self):
        def all_texts():
            tm = texts.TextsMerger()
            tm.perform_merge()

        self.PerformOperation(all_texts, title='Remerging Texts')


    def MergeActors_Clicked(self):
        self.PerformOperation(data.merge_actorinfo, title='Remerging Actor Info')


    def MergeGamedata_Clicked(self):
        def both_datas():
            data.merge_gamedata()
            data.merge_savedata()
        self.PerformOperation(both_datas, title='Remerging Game/Save Data')


    def MergeMaps_Clicked(self):
        def map_and_rstb():
            mubin.merge_maps()
            rstable.generate_master_rstb()
        self.PerformOperation(map_and_rstb, title='Remerging Maps')


    def DeepMerge_Clicked(self):
        self.PerformOperation(merge.deep_merge, title='Redoing Deep Merge')


    def ChangeClicked(self):
        def resort_mods(self):
            all_mergers = [merger() for merger in mergers.get_mergers()]
            remergers = set()
            partials = {}
            mods_to_change = []
            for i in range(self.listWidget.count()):
                mod = self.listWidget.item(i).data(QtCore.Qt.UserRole)
                target_priority = i + 100 if util.get_settings_bool(
                    'load_reverse') else 100 + ((self.listWidget.count() - 1) - i)
                if mod.priority != target_priority:
                    mods_to_change.append(BcmlMod(mod.name, target_priority, mod.path))
                    for merger in all_mergers:
                        if merger.is_mod_logged(mod):
                            remergers.add(merger)
                            if merger.can_partial_remerge():
                                if merger.NAME not in partials:
                                    partials[merger.NAME] = set()
                                partials[merger.NAME] |= set(merger.get_mod_affected(mod))
            for mod in sorted(mods_to_change, key=lambda m: m.priority, reverse=True):
                new_path = util.get_modpack_dir() / util.get_mod_id(mod.name, mod.priority)
                shutil.move(str(mod.path), str(new_path))
                rules_name = 'rules.txt'
                if not Path(str(new_path / rules_name)).exists():
                    rules_name = f'{rules_name}.disable'
                rules = util.RulesParser()
                rules.read(str(new_path / rules_name))
                rules['Definition']['fsPriority'] = str(mod.priority)
                with (new_path / rules_name).open('w', encoding='utf-8') as rf:
                    rules.write(rf)
            for merger in mergers.sort_mergers(remergers):
                if merger.NAME in partials:
                    merger.set_options({'only_these': partials[merger.NAME]})
                merger.perform_merge()
            self.LoadMods()
            install.refresh_cemu_mods()

        self.btnChange.setEnabled(False)
        self.PerformOperation(resort_mods, (self))

    def ExportClicked(self):
        def export(self, output: Path):
            print('Loading files...')
            tmp_dir = util.get_work_dir() / 'tmp_export'
            if tmp_dir.drive != util.get_modpack_dir().drive:
                tmp_dir = Path(util.get_modpack_dir().drive) / 'tmp_bcml_export'
            install.link_master_mod(tmp_dir)
            print('Adding rules.txt...')
            rules_path = tmp_dir / 'rules.txt'
            mods = util.get_installed_mods()
            with rules_path.open('w', encoding='utf-8') as rules:
                rules.writelines([
                    '[Definition]\n',
                    'titleIds = 00050000101C9300,00050000101C9400,00050000101C9500\n',
                    'name = Exported BCML Mod\n',
                    'path = The Legend of Zelda: Breath of the Wild/Mods/Exported BCML\n',
                    f'description = Exported merge of {", ".join([mod.name for mod in mods])}\n',
                    'version = 4\n'
                ])
            if output.suffix == '.bnp' or output.name.endswith('.bnp.7z'):
                print('Exporting BNP...')
                install.create_bnp_mod(
                    mod=tmp_dir,
                    output=output,
                    options={'rstb':{'guess':True}}
                )
            else:
                print('Exporting as graphic pack mod...')
                x_args = [str(util.get_exec_dir() / 'helpers' / '7z.exe'),
                          'a', str(output), f'{str(tmp_dir / "*")}']
                subprocess.run(x_args, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, creationflags=util.CREATE_NO_WINDOW)
            shutil.rmtree(str(tmp_dir))

        file_name = QFileDialog.getSaveFileName(
            self,
            'Save Exported Mod',
            str(Path.home()),
            'BCML Nano Patch Mod (*.bnp; *.bnp.7z);;Graphic Pack Archive (*.zip);;All Files (*)'
        )[0]
        if file_name:
            self.PerformOperation(export, self, Path(file_name))

    def BackupClicked(self):
        backup_name, okay = QtWidgets.QInputDialog.getText(
            self,
            'Backup Mod Configuration',
            'Enter a name for your backup (optional)',
            QtWidgets.QLineEdit.Normal,
            '',
            flags=QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowTitleHint
        )
        if okay:
            self.PerformOperation(install.create_backup, (backup_name), title='Backing Up Mods')

    def RestoreClicked(self):
        backups = [backup.stem for backup in install.get_backups()]
        backup, okay = QtWidgets.QInputDialog.getItem(
            self,
            'Restore Backup',
            'Select the backup to restore',
            backups,
            0,
            False,
            flags=QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowTitleHint
        )
        if okay and backup:
            self.PerformOperation(install.restore_backup, (backup))

    def CemuClicked(self):
        if not self._cemu_exe:
            for file in util.get_cemu_dir().glob('*.exe'):
                if 'cemu' in file.stem.lower():
                    self._cemu_exe = file
        if not self._cemu_exe:
            QtWidgets.QMessageBox.warning(
                self,
                'Error',
                'The Cemu executable could not be found.'
            )
        else:
            subprocess.Popen(
                [
                    str(self._cemu_exe),
                    '-g',
                    str(util.get_game_dir().parent / 'code' / 'U-King.rpx')
                ],
                cwd=str(util.get_cemu_dir())
            )

    def RemoveAllClicked(self):
        def uninstall_everything():
            for folder in [item for item in util.get_modpack_dir().glob('*') if item.is_dir()]:
                print(f'Removing {folder.name}...')
                shutil.rmtree(str(folder))
                print('All BCML mods have been uninstalled!')

        if QtWidgets.QMessageBox.question(
                self,
                'Confirm Uninstall',
                'Are you sure you want to uninstall all of your mods?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        ) == QtWidgets.QMessageBox.Yes:
            self.PerformOperation(uninstall_everything, title='Uninstalling All Mods')

    def UninstallClicked(self):
        def uninstall(mods):
            all_mergers = [merger() for merger in mergers.get_mergers()]
            remergers = []
            partials = {}
            for mod in mods:
                for merger in all_mergers:
                    if merger.is_mod_logged(mod):
                        if merger not in remergers:
                            remergers.append(merger)
                        if merger.can_partial_remerge():
                            if merger.NAME not in partials:
                                partials[merger.NAME] = set()
                            partials[merger.NAME] |= set(merger.get_mod_affected(mod))
            for mod in sorted(mods, key=lambda m: m.priority, reverse=True):
                install.uninstall_mod(mod, wait_merge=True)
            for merger in mergers.sort_mergers(remergers):
                if merger.NAME in partials:
                    merger.set_options({'only_these': partials[merger.NAME]})
                merger.perform_merge()
            install.refresh_cemu_mods()

        if self.listWidget.selectedItems():
            mods = [item.data(Qt.UserRole)
                    for item in self.listWidget.selectedItems()]
            if QtWidgets.QMessageBox.question(
                self,
                'Confirm Uninstall',
                f'Are you sure you want to uninstall the selected {len(mods)} mod(s)?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            ) == QtWidgets.QMessageBox.Yes:
                self.PerformOperation(uninstall, (mods))

    def DisableModClicked(self):
        def disable(mods):
            all_mergers = [merger() for merger in mergers.get_mergers()]
            remergers = []
            partials = {}
            for mod in mods:
                for merger in all_mergers:
                    if merger.is_mod_logged(mod):
                        if merger not in remergers:
                            remergers.append(merger)
                        if merger.can_partial_remerge():
                            if merger.NAME not in partials:
                                partials[merger.NAME] = set()
                            partials[merger.NAME] |= set(merger.get_mod_affected(mod))
            for mod in mods:
                install.disable_mod(mod, wait_merge=True)
            for merger in mergers.sort_mergers(remergers):
                if merger.NAME in partials:
                    merger.set_options({'only_these': partials[merger.NAME]})
                merger.perform_merge()
            install.refresh_cemu_mods()

        if self.listWidget.selectedItems():
            mods = [item.data(Qt.UserRole)
                    for item in self.listWidget.selectedItems()]
            if QtWidgets.QMessageBox.question(
                self,
                'Confirm Disable',
                f'Are you sure you want to disable the selected {len(mods)} mod(s)?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            ) == QtWidgets.QMessageBox.Yes:
                self.PerformOperation(disable, (mods))

    def EnableModClicked(self):
        def enable(mods):
            all_mergers = [merger() for merger in mergers.get_mergers()]
            remergers = []
            partials = {}
            for mod in mods:
                for merger in all_mergers:
                    if merger.is_mod_logged(mod):
                        if merger not in remergers:
                            remergers.append(merger)
                        if merger.can_partial_remerge():
                            if merger.NAME not in partials:
                                partials[merger.NAME] = set()
                            partials[merger.NAME] |= set(merger.get_mod_affected(mod))
            for mod in mods:
                install.enable_mod(mod, wait_merge=True)
            for merger in mergers.sort_mergers(remergers):
                if merger.NAME in partials:
                    merger.set_options({'only_these': partials[merger.NAME]})
                merger.perform_merge()
            install.refresh_cemu_mods()

        if self.listWidget.selectedItems():
            mods = [item.data(Qt.UserRole)
                    for item in self.listWidget.selectedItems()]
            self.PerformOperation(enable, (mods))

    def ExploreClicked(self):
        path = self.listWidget.selectedItems()[0].data(QtCore.Qt.UserRole).path
        if platform.system() == "Windows":
            os.startfile(path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", path])
        else:
            subprocess.Popen(["xdg-open", path])

    def PackageClicked(self):
        def create_package(result):
            rules = [
                '[Definition]',
                'titleIds = 00050000101C9300,00050000101C9400,00050000101C9500',
                f'name = {result["name"]}',
                f'path = {{BCML: DON\'T TOUCH}}/{result["name"]}',
                'description = {}'.format(result['description'].replace('\n', ' ')),
                'version = 4'
            ]
            if result['image']:
                rules.append(f'image = {result["image"]}')
            if result['url']:
                rules.append(f'url = {result["url"]}')
            (result['folder'] / 'rules.txt').write_text('\n'.join(rules), encoding='utf-8')
            install.create_bnp_mod(
                mod=result['folder'],
                output=result['output'],
                options=result['options']
            )
        dialog = PackageDialog(self)
        result = dialog.GetResult()
        if result:
            self.PerformOperation(
                create_package,
                (result),
                title=f'Creating BNP package for {result["name"]}'
            )

    def SettingsClicked(self):
        if SettingsDialog(self).exec_() == QtWidgets.QDialog.Accepted:
            self.LoadMods()

    def AboutClicked(self):
        AboutDialog(self).exec_()

# Install Dialog


class InstallDialog(QtWidgets.QDialog, Ui_InstallDialog):

    def __init__(self, *args, **kwargs): # pylint: disable=unused-argument
        self._options = {
            'disable': [],
            'options': {}
        }
        super(InstallDialog, self).__init__()
        self.setupUi(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)
        self.btnAddFile.clicked.connect(self.AddFileClicked)
        self.btnAddFolder.clicked.connect(self.AddFolderClicked)
        self.btnRemove.clicked.connect(self.RemoveClicked)
        self.btnClear.clicked.connect(self.ClearClicked)
        self.btnAdvanced.clicked.connect(self.AdvancedClicked)
        self.spnPriority.setValue(install.get_next_priority())

    def AddFileClicked(self):
        file_names = QFileDialog.getOpenFileNames(
            self,
            'Select a Mod',
            str(Path.home()),
            'Mod Archives (*.zip; *.rar; *.7z; *.bnp);;All Files (*)'
        )[0]
        for file_name in file_names:
            path = Path(file_name)
            if path.exists():
                mod_item = QtWidgets.QListWidgetItem()
                mod_item.setText(path.stem)
                mod_item.setData(Qt.UserRole, path)
                self.lstQueue.addItem(mod_item)

    def AddFolderClicked(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'Select a Mod Folder')
        if folder:
            path = Path(folder)
            if path.exists():
                mod_item = QtWidgets.QListWidgetItem()
                mod_item.setText(path.stem)
                mod_item.setData(Qt.UserRole, path)
                self.lstQueue.addItem(mod_item)

    def RemoveClicked(self):
        for i in self.lstQueue.selectedIndexes():
            self.lstQueue.takeItem(i.row())

    def ClearClicked(self):
        self.lstQueue.clear()

    def AdvancedClicked(self):
        self._options = OptionsDialog(self).get_options()

    def GetResult(self):
        if self.exec_() == QtWidgets.QDialog.Accepted:
            paths = [self.lstQueue.item(i).data(Qt.UserRole)
                     for i in range(self.lstQueue.count())]
            return {
                'paths': paths,
                'insert_priority': self.spnPriority.value(),
                'options': self._options
            }

# Options Dialog


class OptionsDialog(QtWidgets.QDialog, Ui_OptionsDialog):

    def __init__(self, *args, **kwargs):
        super(OptionsDialog, self).__init__()
        self.setupUi(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)

    def get_options(self) -> dict:
        if self.exec_():
            options = {
                'disable': []
            }
            for widget in self.checkboxes:
                if hasattr(widget, 'disable_name') and widget.isChecked():
                    options['disable'].append(widget.disable_name)
                if hasattr(widget, 'option_name'):
                    if not widget.merger in options:
                        options[widget.merger] = {}
                    options[widget.merger][widget.option_name] = widget.isChecked()
            return options
        else:
            return {}

# Package Dialog


class PackageDialog(QtWidgets.QDialog, Ui_PackageDialog):

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        super(PackageDialog, self).__init__()
        self.setupUi(self)
        self._options = {}
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)
        self.btnBrowseContent.clicked.connect(self.BrowseContentClicked)
        self.btnBrowseImg.clicked.connect(self.BrowseImgClicked)
        self.btnAdvanced.clicked.connect(self.AdvancedClicked)

    def BrowseContentClicked(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'Select Mod Folder')
        if folder:
            path = Path(folder)
            if path.name == 'content':
                path = path.parent
            self.txtFolder.setText(str(path.resolve()))
            if (path / 'rules.txt').exists():
                rules = util.RulesParser()
                rules.read(str(path / 'rules.txt'))
                if 'name' in rules['Definition'] and not self.txtName.text():
                    self.txtName.setText(str(rules['Definition']['name']))
                if 'url' in rules['Definition'] and not self.txtUrl.text():
                    self.txtUrl.setText(str(rules['Definition']['url']))
                if 'image' in rules['Definition'] and not self.txtImage.text():
                    self.txtImage.setText(str(rules['Definition']['image']))
                if 'description' in rules['Definition'] and not self.txtDescript.toPlainText():
                    self.txtDescript.setPlainText(str(rules['Definition']['description']))

    def BrowseImgClicked(self):
        file_name = QFileDialog.getOpenFileName(
            self,
            'Select a Preview Image',
            str(Path.home()),
            'Images (*.png; *.jpg; *.bmp);;All Files (*)'
        )[0]
        if file_name:
            self.txtImage.setText(str(Path(file_name).resolve()))

    def AdvancedClicked(self):
        self._options = OptionsDialog(self).get_options()

    def accept(self):
        if not self.txtName.text().strip():
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'You must provide a name for your mod.')
            return
        if not self.txtDescript.toPlainText().strip():
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'You must provide a description for your mod.')
            return
        if not self.txtFolder.text().strip() or \
                not Path(self.txtFolder.text().strip()).exists():
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'You must provide a valid folder containing a mod.')
            return
        return super().accept()

    def GetResult(self):
        if self.exec_() == QtWidgets.QDialog.Accepted:
            output = QFileDialog.getSaveFileName(
                self,
                'Save Your Mod',
                str(Path.home()),
                'Botw Nano Patch Mod (*.bnp);;7-Zip Archive (*.7z);;All Files (*)'
            )[0]
            if not output:
                return None
            return {
                'name': self.txtName.text().strip(),
                'description': self.txtDescript.toPlainText().strip(),
                'image': self.txtImage.text().strip(),
                'url': self.txtUrl.text().strip(),
                'folder': Path(self.txtFolder.text().strip()),
                'options': self._options,
                'output': Path(output)
            }


# Settings Dialog

class SettingsDialog(QtWidgets.QDialog, Ui_SettingsDialog):

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        super(SettingsDialog, self).__init__()
        self.setupUi(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(
            str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)

        self.txtCemu.setText(str(util.get_cemu_dir()))
        self.txtGameDump.setText(str(util.get_game_dir()))
        try:
            self.txtMlc.setText(str(util.get_mlc_dir()))
        except FileNotFoundError:
            self.txtMlc.setText('')
        self.chkDark.setChecked(util.get_settings_bool('dark_theme'))
        self.chkGuessMerge.setChecked(util.get_settings_bool('guess_merge'))
        self.drpLang.addItems(texts.LANGUAGES)
        self.drpLang.setCurrentText(util.get_settings()['lang'])

        self.btnBrowseCemu.clicked.connect(self.BrowseCemuClicked)
        self.btnBrowseGame.clicked.connect(self.BrowseGameClicked)
        self.btnBrowseMlc.clicked.connect(self.BrowseMlcClicked)

    def BrowseCemuClicked(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'Select Cemu Directory')
        if folder:
            self.txtCemu.setText(folder)

    def BrowseGameClicked(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'Select Game Dump Directory')
        if folder:
            self.txtGameDump.setText(folder)

    def BrowseMlcClicked(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'Select Cemu MLC Directory')
        if folder:
            self.txtMlc.setText(folder)

    def accept(self):
        cemu_path = Path(self.txtCemu.text())
        if cemu_path.exists():
            util.set_cemu_dir(cemu_path)
        else:
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'The specified Cemu directory does not exist.')
            return
        game_path = Path(self.txtGameDump.text())
        if game_path.exists():
            util.set_game_dir(game_path)
        else:
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'The specified game dump directory does not exist.')
            return
        mlc_path = Path(self.txtMlc.text())
        if mlc_path.exists():
            util.set_mlc_dir(mlc_path)
        else:
            QtWidgets.QMessageBox.warning(
                self, 'Error', 'The specified Cemu MLC directory does not exist.')
            return
        dark = util.get_settings_bool('dark_theme')
        if dark != self.chkDark.isChecked():
            util.set_settings_bool('dark_theme', self.chkDark.isChecked())
            if self.chkDark.isChecked():
                QtWidgets.QApplication.instance().setStyleSheet(DARK_THEME)
            else:
                QtWidgets.QApplication.instance().setStyleSheet('')
        util.set_settings_bool('guess_merge', self.chkGuessMerge.isChecked())
        util.get_settings()['lang'] = self.drpLang.currentText()
        util.save_settings()
        return super().accept()

# About Dialog


class AboutDialog(QtWidgets.QDialog, Ui_AboutDialog):

    def link(self, linkStr):
        QDesktopServices.openUrl(QUrl(linkStr))

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        super(AboutDialog, self).__init__()
        self.setupUi(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(
            str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)
        logo_path = str(util.get_exec_dir() / 'data' / 'logo-smaller.png')
        self.lblLogo.setPixmap(QtGui.QPixmap(logo_path).scaledToWidth(256))
        self.lblVersion.setText('Version ' + util.get_bcml_version())
        self.lblSites.linkActivated.connect(self.link)
        self.btnClose.clicked.connect(self.close)

# Progress Operation Stuff


class ProgressDialog(QtWidgets.QDialog, Ui_ProgressDialog):

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        super(ProgressDialog, self).__init__()
        self.setupUi(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(
            str(util.get_exec_dir() / 'data' / 'bcml.ico')))
        self.setWindowIcon(icon)
        self.setWindowFlags(QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowTitleHint)


class ProgressThread(threading.Thread):
    def __init__(self, output, target, *args):
        threading.Thread.__init__(self)
        self._out = output
        self._target = target
        self._args = args
        self.signal = ThreadSignal()
        self.error = None

    def run(self):
        sys.stdout = RedirectText(self._out)
        try:
            self._target(*self._args)
            self.signal.sig.emit('Done')
        except Exception as e: # pylint: disable=broad-except
            if hasattr(e, 'error_text'):
                self.error = e.error_text # pylint: disable=no-member
            else:
                self.error = traceback.format_exc(limit=-4)
            self.signal.err.emit(e)


class RedirectText:
    def __init__(self, label: QtWidgets.QLabel):
        self.out = label

    def write(self, text):
        if text != '\n':
            try:
                self.out.setText(text)
            except RuntimeError:
                pass

    def flush(self):
        pass


class ThreadSignal(QtCore.QObject):
    sig = QtCore.Signal(str)
    err = QtCore.Signal(str)


DARK_THEME = """
    QMainWindow, QDialog {
        background-color: #2f3136;
    }

    QMainWindow, QLabel, QGroupBox, QCheckBox {
        color: #eeeeee;
    }

    QStatusBar, QStatusBar QLabel {
        color: #7f8186;
    }

    QStatusBar:item {
        border: 0;
    }

    QPushButton, QToolButton {
        background-color: #1584CD;
        color: #e0e0e0;
    }

    QToolButton {
        padding: 0px 2px;
        border-radius: 2px;
    }

    QPushButton {
        border-radius: 2px;
        padding: 4px 6px;
    }

    QPushButton:disabled, QToolButton:disabled {
        background-color: #4f5b62;
        color: #212121;
    }

    QPushButton:hover, QToolButton:hover {
        background-color: #3899D8;
    }

    QPushButton:focus, QToolButton:focus {
        background-color: #0765A3;
    }

    QGroupBox {
        background-color: #212121;
        border: 1px solid #202225;
        border-radius: 4px;
        padding-top: 8px;
        margin-top: 6px;
    }

    QGroupBox:title {
        subcontrol-origin: margin;
        top: 0;
        left: 12px;
        padding: 0 4px;
    }

    QListWidget {
        background-color: #212121;
        border: 1px solid #202225;
        border-radius: 2px;
        color: #fafafa;
        padding: 0 4px;
    }

    QListWidget:item {
        padding: 4px 0;
    }

    QListWidget:item:alternate {
        background-color: #424242;
    }

    QLineEdit, QPlainTextEdit {
        background-color: #484c52;
        border: 0;
        border-radius: 2px;
        color: #fafafa;
        padding: 2px;
    }    

    QScrollBar:vertical {
        background-color: #212121;
        border: 0;
        width: 10px;
        margin: 0;
    }

    QScrollBar::handle:vertical {
        background: #424242;
        min-height: 0;
    }

    QScrollBar::add-line:vertical {
        border: 0;
        background: #32CC99;
        height: 0;
        subcontrol-position: bottom;
        subcontrol-origin: margin;
    }

    QScrollBar::sub-line:vertical {
        border: 0;
        background: #32CC99;
        height: 0;
        subcontrol-position: top;
        subcontrol-origin: margin;
    }

    QScrollBar::up-arrow:vertical, QScrollBar::down-arrow:vertical {
        border: 0;
        width: 3px;
        height: 0;
        background: white;
    }

    QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical {
        background: none;
    }

    QMenu {
        background-color: #212121;
        color: #fafafa;
    }

    QMenu:item:selected {
        background-color: #424242;
    }
"""

# Main

def uri_validator(x):
    from urllib.parse import urlparse
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc, result.path])
    except Exception: # pylint: disable=broad-except
        return False


def quit_download():
    del process_args.progress
    raise RuntimeError()


def process_args() -> Path:
    try:
        if Path(sys.argv[1]).exists():
            return Path(sys.argv[1])
    except (WindowsError, OSError):
        if sys.argv[1].startswith('bcml:'):
            try_url = sys.argv[1].replace('bcml:', '').replace('\\', '/')
            if uri_validator(try_url) and 'gamebanana.com' in try_url:
                process_args.progress = QtWidgets.QProgressDialog(
                    'Downloading requested mod...', 'Stop', 0, 0, None
                )
                from tempfile import NamedTemporaryFile
                try:
                    with NamedTemporaryFile('wb', prefix='GameBanana_',
                                            suffix='.bnp', delete=False) as tmp:
                        path = Path(tmp.name)
                    process_args.progress.open()
                    process_args.progress.canceled.connect(quit_download)
                    urllib.request.urlretrieve(try_url, path.resolve(), reporthook=download_progress)
                    process_args.progress.canceled.disconnect()
                    process_args.progress.cancel()
                    return Path(tmp.name)
                except RuntimeError:
                    if hasattr(process_args, 'progress') and process_args.progress:
                        process_args.progress.cancel()
                except Exception as e: # pylint: disable=broad-except
                    if hasattr(process_args, 'progress') and process_args.progress:
                        process_args.progress.cancel()
                    QtWidgets.QMessageBox.warning(
                        None,
                        'Error',
                        f'Failed to download mod from <code>{try_url}</code>. '
                        f'Error details:\n\n{str(e)}'
                    )
            else:
                QtWidgets.QMessageBox.warning(
                    None,
                    'Error',
                    f'The request download <code>{try_url}</code> does not appear to be a valid '
                    'GameBanana URL.'
                )
    return None


def download_progress(count, block_size, total_size):
    QtCore.QCoreApplication.instance().processEvents()
    process_args.progress.setMaximum(total_size)
    process_args.progress.setValue(count * block_size)


def main():
    util.clear_temp_dir()
    util.create_schema_handler()
    app = QtWidgets.QApplication([])
    ver = platform.python_version_tuple()
    if int(ver[0]) < 3 or (int(ver[0]) >= 3 and int(ver[1]) < 7):
        QtWidgets.QMessageBox.warning(
            None,
            'Error',
            f'BCML requires Python 3.7 or higher, but your Python version is {ver[0]}.{ver[1]}'
        )
        sys.exit(0)

    is_64bits = sys.maxsize > 2**32
    if not is_64bits:
        QtWidgets.QMessageBox.warning(
            None,
            'Error',
            'BCML requires 64 bit Python, but it looks like you\'re running 32 bit.'
        )
        sys.exit(0)
    if util.get_settings_bool('dark_theme'):
        app.setStyleSheet(DARK_THEME)
        if 'Roboto Lt' in QtGui.QFontDatabase().families(QtGui.QFontDatabase.Latin):
            app.setFont(QtGui.QFont('Roboto Lt', 10, weight=QtGui.QFont.DemiBold))
    application = MainWindow()
    try:
        application.show()
        application.SetupChecks()
        if len(sys.argv) > 1:
            parg = process_args()
            if parg:
                application.InstallClicked(parg)
        app.exec_()
    except Exception: # pylint: disable=broad-except
        tb = traceback.format_exc(limit=-2)
        e = util.get_work_dir() / 'error.log'
        QtWidgets.QMessageBox.warning(
            None,
            'Error',
            f'An unexpected error has occured:\n\n{tb}\n\nThe error has been logged to:\n'
            f'{e}\n\nBCML will now close.'
        )
        e.write_text(tb, encoding='utf-8')
    finally:
        util.clear_temp_dir()


if __name__ == "__main__":
    main()



"""Provides functions for diffing and merging SARC packs"""
# Copyright 2019 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
import csv
import io
import os
from functools import partial
from multiprocessing import Pool, cpu_count
from pathlib import Path
from typing import List, Union

import sarc
import xxhash

from bcml import data, util, mergers
from bcml.util import BcmlMod


def get_pack_mods() -> List[BcmlMod]:
    """
    Gets a list of all installed pack mods

    :return: Returns a list of mods that modify pack files
    :rtype: list of class:`bcml.util.BcmlMod`
    """
    pmods = [mod for mod in util.get_installed_mods() if (
        mod.path / 'logs' / 'packs.log').exists()]
    return sorted(pmods, key=lambda mod: mod.priority)


def get_modded_packs_in_mod(mod: Union[Path, str, BcmlMod]) -> List[str]:
    """
    Get all pack files modified by a given mod
    """
    path = mod if isinstance(mod, Path) else Path(
        mod) if isinstance(mod, str) else mod.path
    packs = []
    plog = path / 'logs' / 'packs.log'
    if not plog.exists():
        return []
    with plog.open('r') as rlog:
        csv_loop = csv.reader(rlog)
        for row in csv_loop:
            if 'logs' in str(row[1]) or str(row[0]) == 'name':
                continue
            packs.append(Path(str(row[1])).as_posix())
    return packs


def merge_sarcs(file_name: str, sarcs: List[Union[Path, bytes]]) -> (str, bytes):
    opened_sarcs: List[sarc.SARC] = []
    if isinstance(sarcs[0], Path):
        for i, sarc_path in enumerate(sarcs):
            sarcs[i] = sarc_path.read_bytes()
    for sarc_bytes in sarcs:
        sarc_bytes = util.unyaz_if_needed(sarc_bytes)
        try:
            opened_sarcs.append(sarc.SARC(sarc_bytes))
        except ValueError:
            continue

    all_files = {key for open_sarc in opened_sarcs for key in open_sarc.list_files()}
    nested_sarcs = {}
    new_sarc = sarc.SARCWriter(be=True)
    files_added = []

    # for file in all_files:
    #     dm_cache = util.get_master_modpack_dir() / 'logs' / 'dm' / file
    #     if dm_cache.exists():
    #         file_data = dm_cache.read_bytes()
    #         new_sarc.add_file(file, file_data)
    #         files_added.append(file)

    for opened_sarc in reversed(opened_sarcs):
        for file in [file for file in opened_sarc.list_files() if file not in files_added]:
            data = opened_sarc.get_file_data(file).tobytes()
            if util.is_file_modded(file.replace('.s', '.'), data, count_new=True):
                if not Path(file).suffix in util.SARC_EXTS:
                    new_sarc.add_file(file, data)
                    files_added.append(file)
                else:
                    if file not in nested_sarcs:
                        nested_sarcs[file] = []
                    nested_sarcs[file].append(util.unyaz_if_needed(data))
    for file, sarcs in nested_sarcs.items():
        merged_bytes = merge_sarcs(file, sarcs)[1]
        if Path(file).suffix.startswith('.s') and not file.endswith('.sarc'):
            merged_bytes = util.compress(merged_bytes)
        new_sarc.add_file(file, merged_bytes)
        files_added.append(file)
    for file in [file for file in all_files if file not in files_added]:
        for opened_sarc in [open_sarc for open_sarc in opened_sarcs \
                            if file in open_sarc.list_files()]:
            new_sarc.add_file(file, opened_sarc.get_file_data(file).tobytes())
            break

    if 'Bootup.pack' in file_name:
        for merger in [merger() for merger in mergers.get_mergers() if merger.is_bootup_injector()]:
            inject = merger.get_bootup_injection()
            if not inject:
                continue
            file, data = inject
            try:
                new_sarc.delete_file(file)
            except KeyError:
                pass
            new_sarc.add_file(file, data)

    return (file_name, new_sarc.get_bytes())


class PackMerger(mergers.Merger):
    """ A merger for modified pack files """
    NAME: str = 'packs'

    def __init__(self):
        super().__init__('SARC pack merger', 'Merges modified files within SARCs', 'packs.log', {})

    def can_partial_remerge(self):
        return True

    def get_mod_affected(self, mod):
        return get_modded_packs_in_mod(mod)

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[str, Path]]):
        packs = {}
        for file in [file for file in modded_files \
                     if isinstance(file, Path) and file.suffix in util.SARC_EXTS]:
            canon = util.get_canon_name(file.relative_to(mod_dir).as_posix())
            if canon and not any(ex in file.name for ex in ['Dungeon', 'Bootup_', 'AocMainField']):
                packs[canon] = file.relative_to(mod_dir).as_posix()
        return packs

    def log_diff(self, mod_dir: Path, diff_material: Union[dict, List[Path]]):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(mod_dir, diff_material)
        with (mod_dir / 'logs' / self._log_name).open('w', encoding='utf-8') as log:
            log.write('name,path\n')
            for canon, path in diff_material.items():
                if 'logs' in path:
                    continue
                log.write(f'{canon},{path}\n')

    def get_mod_diff(self, mod: util.BcmlMod):
        return get_modded_packs_in_mod(mod)

    def get_all_diffs(self):
        diffs = {}
        for mod in [mod for mod in util.get_installed_mods() if self.is_mod_logged(mod)]:
            diffs[mod] = get_modded_packs_in_mod(mod)
        return diffs

    def consolidate_diffs(self, diffs):
        all_sarcs = set()
        all_diffs = {}
        for mod in sorted(diffs.keys(), key=lambda mod: mod.priority):
            all_sarcs |= set(diffs[mod])
        for modded_sarc in all_sarcs:
            for mod, diff in diffs.items():
                if modded_sarc in diff:
                    if not modded_sarc in all_diffs:
                        all_diffs[modded_sarc] = []
                    if (mod.path / modded_sarc).exists():
                        all_diffs[modded_sarc].append(mod.path / modded_sarc)
        return all_diffs

    def perform_merge(self):
        print('Loading modded SARC list...')
        sarcs = self.consolidate_diffs(self.get_all_diffs())
        if 'only_these' in self._options:
            for sarc_file in self._options['only_these']:
                master_path = (util.get_master_modpack_dir() / sarc_file)
                if master_path.exists():
                    master_path.unlink()
            for sarc_file in [file for file in sarcs if file not in self._options['only_these']]:
                del sarcs[sarc_file]
        else:
            for file in [file for file in util.get_master_modpack_dir().rglob('**/*') \
                        if file.suffix in util.SARC_EXTS]:
                file.unlink()
        for sarc_file in sarcs:
            try:
                sarcs[sarc_file].insert(0, util.get_game_file(sarc_file))
            except FileNotFoundError:
                continue
        if not sarcs:
            print('No SARC merging necessary')
            return
        num_threads = min(cpu_count(), len(sarcs))
        pool = Pool(processes=num_threads)
        print(f'Merging {len(sarcs)} SARC files...')
        results = pool.starmap(merge_sarcs, sarcs.items())
        pool.close()
        pool.join()
        for result in results:
            file, data = result
            output_path = util.get_master_modpack_dir() / file
            output_path.parent.mkdir(parents=True, exist_ok=True)
            if output_path.suffix.startswith('.s'):
                data = util.compress(data)
            output_path.write_bytes(data)
        print('Finished merging SARCs')

    def get_checkbox_options(self):
        return []



# pylint: skip-file

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_ProgressDialog(object):
    def setupUi(self, ProgressDialog):
        ProgressDialog.setObjectName("ProgressDialog")
        ProgressDialog.resize(400, 80)
        ProgressDialog.setMinimumSize(QtCore.QSize(400, 80))
        ProgressDialog.setMaximumSize(QtCore.QSize(400, 80))
        self.verticalLayout = QtWidgets.QVBoxLayout(ProgressDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.progressBar = QtWidgets.QProgressBar(ProgressDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progressBar.sizePolicy().hasHeightForWidth())
        self.progressBar.setSizePolicy(sizePolicy)
        self.progressBar.setMaximum(0)
        self.progressBar.setProperty("value", -1)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout.addWidget(self.progressBar)
        self.lblProgress = QtWidgets.QLabel(ProgressDialog)
        self.lblProgress.setObjectName("lblProgress")
        self.verticalLayout.addWidget(self.lblProgress)

        self.retranslateUi(ProgressDialog)
        QtCore.QMetaObject.connectSlotsByName(ProgressDialog)

    def retranslateUi(self, ProgressDialog):
        ProgressDialog.setWindowTitle(QtWidgets.QApplication.translate("ProgressDialog", "BCML Operation in Progress", None, -1))
        self.lblProgress.setText(QtWidgets.QApplication.translate("ProgressDialog", "...", None, -1))




# pylint: skip-file

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(640, 480)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setSpacing(8)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(8)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setObjectName("label")
        self.horizontalLayout_4.addWidget(self.label)
        self.btnOrder = QtWidgets.QToolButton(self.centralwidget)
        self.btnOrder.setObjectName("btnOrder")
        self.horizontalLayout_4.addWidget(self.btnOrder)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.listWidget = QtWidgets.QListWidget(self.centralwidget)
        self.listWidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout_2.addWidget(self.listWidget)
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setSpacing(4)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.btnExport = QtWidgets.QPushButton(self.centralwidget)
        self.btnExport.setEnabled(False)
        self.btnExport.setObjectName("btnExport")
        self.gridLayout_3.addWidget(self.btnExport, 0, 3, 1, 1)
        self.btnInstall = QtWidgets.QPushButton(self.centralwidget)
        self.btnInstall.setObjectName("btnInstall")
        self.gridLayout_3.addWidget(self.btnInstall, 0, 0, 1, 1)
        self.btnRemerge = QtWidgets.QToolButton(self.centralwidget)
        self.btnRemerge.setStyleSheet("padding: 3px 10px 2px;")
        self.btnRemerge.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.btnRemerge.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.btnRemerge.setArrowType(QtCore.Qt.NoArrow)
        self.btnRemerge.setObjectName("btnRemerge")
        self.gridLayout_3.addWidget(self.btnRemerge, 0, 1, 1, 1)
        self.btnChange = QtWidgets.QPushButton(self.centralwidget)
        self.btnChange.setEnabled(False)
        self.btnChange.setObjectName("btnChange")
        self.gridLayout_3.addWidget(self.btnChange, 0, 2, 1, 1)
        self.btnBackup = QtWidgets.QPushButton(self.centralwidget)
        self.btnBackup.setObjectName("btnBackup")
        self.gridLayout_3.addWidget(self.btnBackup, 1, 0, 1, 1)
        self.btnRestore = QtWidgets.QPushButton(self.centralwidget)
        self.btnRestore.setObjectName("btnRestore")
        self.gridLayout_3.addWidget(self.btnRestore, 1, 1, 1, 1)
        self.btnRemoveAll = QtWidgets.QPushButton(self.centralwidget)
        self.btnRemoveAll.setObjectName("btnRemoveAll")
        self.gridLayout_3.addWidget(self.btnRemoveAll, 1, 2, 1, 1)
        self.btnCemu = QtWidgets.QPushButton(self.centralwidget)
        self.btnCemu.setObjectName("btnCemu")
        self.gridLayout_3.addWidget(self.btnCemu, 1, 3, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout_3)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox.setMaximumSize(QtCore.QSize(276, 16777215))
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_2.setContentsMargins(-1, 3, -1, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setContentsMargins(-1, -1, -1, 8)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lblImage = QtWidgets.QLabel(self.groupBox)
        self.lblImage.setText("")
        self.lblImage.setScaledContents(True)
        self.lblImage.setObjectName("lblImage")
        self.verticalLayout.addWidget(self.lblImage)
        self.lblModInfo = QtWidgets.QLabel(self.groupBox)
        self.lblModInfo.setObjectName("lblModInfo")
        self.verticalLayout.addWidget(self.lblModInfo)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.btnUninstall = QtWidgets.QPushButton(self.groupBox)
        self.btnUninstall.setEnabled(False)
        self.btnUninstall.setObjectName("btnUninstall")
        self.horizontalLayout_2.addWidget(self.btnUninstall)
        self.btnDisableMod = QtWidgets.QPushButton(self.groupBox)
        self.btnDisableMod.setEnabled(False)
        self.btnDisableMod.setObjectName("btnDisableMod")
        self.horizontalLayout_2.addWidget(self.btnDisableMod)
        self.btnEnableMod = QtWidgets.QPushButton(self.groupBox)
        self.btnEnableMod.setEnabled(False)
        self.btnEnableMod.setVisible(False)
        self.btnEnableMod.setObjectName("btnEnableMod")
        self.horizontalLayout_2.addWidget(self.btnEnableMod)
        self.btnExplore = QtWidgets.QPushButton(self.groupBox)
        self.btnExplore.setEnabled(False)
        self.btnExplore.setObjectName("btnExplore")
        self.horizontalLayout_2.addWidget(self.btnExplore)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.verticalLayout.setStretch(2, 6)
        self.gridLayout_2.addLayout(self.verticalLayout, 1, 0, 1, 1)
        self.horizontalLayout.addWidget(self.groupBox)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setStyleSheet("margin-right: 8px;")
        self.statusbar.setSizeGripEnabled(False)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionInstallFolder = QtWidgets.QAction(MainWindow)
        self.actionInstallFolder.setIconVisibleInMenu(False)
        self.actionInstallFolder.setObjectName("actionInstallFolder")
        self.actionInstallMultiple = QtWidgets.QAction(MainWindow)
        self.actionInstallMultiple.setIconVisibleInMenu(False)
        self.actionInstallMultiple.setObjectName("actionInstallMultiple")

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QtWidgets.QApplication.translate("MainWindow", "BCML: BotW Cemu Mod Loader", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("MainWindow", "Installed Mods", None, -1))
        self.btnOrder.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Toggle the order in which mods are displayed.\nDefault lists mods from highest priority to lowest priority, but can be reversed.", None, -1))
        self.btnOrder.setText(QtWidgets.QApplication.translate("MainWindow", "...", None, -1))
        self.listWidget.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Drag and drop to change mod load order. Mods at the top of the list override mods at the bottom.", None, -1))
        self.btnInstall.setText(QtWidgets.QApplication.translate("MainWindow", "Install...", None, -1))
        self.btnRemerge.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Perform all merges again", None, -1))
        self.btnRemerge.setText(QtWidgets.QApplication.translate("MainWindow", "Remerge", None, -1))
        self.btnChange.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Apply changes to load order", None, -1))
        self.btnChange.setText(QtWidgets.QApplication.translate("MainWindow", " Apply Sort", None, -1))
        self.btnExport.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Export installed mods as a single mod", None, -1))
        self.btnExport.setText(QtWidgets.QApplication.translate("MainWindow", "Export...", None, -1))
        self.btnBackup.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Creates a backup of all currently installed mods", None, -1))
        self.btnBackup.setText(QtWidgets.QApplication.translate("MainWindow", "Backup Mods", None, -1))
        self.btnRestore.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Restores a mod configuration backup", None, -1))
        self.btnRestore.setText(QtWidgets.QApplication.translate("MainWindow", "Restore...", None, -1))
        self.btnRemoveAll.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Uninstalls all mods and clears BCML's master pack", None, -1))
        self.btnRemoveAll.setText(QtWidgets.QApplication.translate("MainWindow", "Uninstall All", None, -1))
        self.btnCemu.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Open Cemu and run Breath of the Wild", None, -1))
        self.btnCemu.setText(QtWidgets.QApplication.translate("MainWindow", "Launch Game", None, -1))
        self.groupBox.setTitle(QtWidgets.QApplication.translate("MainWindow", "Mod Info", None, -1))
        self.lblModInfo.setText(QtWidgets.QApplication.translate("MainWindow", "No mod selected", None, -1))
        self.btnUninstall.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Uninstall the selected mod", None, -1))
        self.btnUninstall.setText(QtWidgets.QApplication.translate("MainWindow", "Uninstall", None, -1))
        self.btnDisableMod.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Disable the selected mod", None, -1))
        self.btnDisableMod.setText(QtWidgets.QApplication.translate("MainWindow", "Disable", None, -1))
        self.btnEnableMod.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Enable the selected mod", None, -1))
        self.btnEnableMod.setText(QtWidgets.QApplication.translate("MainWindow", "Enable", None, -1))
        self.btnExplore.setToolTip(QtWidgets.QApplication.translate("MainWindow", "Open the selected mod in your default file manager", None, -1))
        self.btnExplore.setText(QtWidgets.QApplication.translate("MainWindow", "Explore...", None, -1))
        self.actionInstallFolder.setText(QtWidgets.QApplication.translate("MainWindow", "Install from Folder", None, -1))
        self.actionInstallMultiple.setText(QtWidgets.QApplication.translate("MainWindow", "Install Multiple", None, -1))




"""
Provides functions to diff and merge various kinds of BotW data, including gamedata, savedata, and
actorinfo.
"""
# Copyright 2019 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
import zlib
from copy import deepcopy
from functools import partial
from io import BytesIO
from math import ceil
from multiprocessing import Pool, cpu_count
from pathlib import Path
from typing import List, Union

import byml
from byml import yaml_util
import rstb
import rstb.util
import sarc
import xxhash
import yaml

from bcml import util, rstable, mergers
from bcml.util import BcmlMod


def get_stock_gamedata() -> sarc.SARC:
    """ Gets the contents of the unmodded gamedata.sarc """
    if not hasattr(get_stock_gamedata, 'gamedata'):
        with util.get_game_file('Pack/Bootup.pack').open('rb') as b_file:
            bootup = sarc.read_file_and_make_sarc(b_file)
        get_stock_gamedata.gamedata = sarc.SARC(util.decompress(
            bootup.get_file_data('GameData/gamedata.ssarc')))
    return get_stock_gamedata.gamedata


def get_stock_savedata() -> sarc.SARC:
    """ Gets the contents of the unmodded savedataformat.sarc """
    if not hasattr(get_stock_savedata, 'savedata'):
        with util.get_game_file('Pack/Bootup.pack').open('rb') as b_file:
            bootup = sarc.read_file_and_make_sarc(b_file)
        get_stock_savedata.savedata = sarc.SARC(util.decompress(
            bootup.get_file_data('GameData/savedataformat.ssarc')
        ))
    return get_stock_savedata.savedata


def get_gamedata_hashes() -> {}:
    """ Gets a hash table for the unmodded contents of gamedata.sarc """
    if not hasattr(get_gamedata_hashes, 'gamedata_hashes'):
        get_gamedata_hashes.gamedata_hashes = {}
        gamedata = get_stock_gamedata()
        for file in gamedata.list_files():
            get_gamedata_hashes.gamedata_hashes[file] = xxhash.xxh32(
                gamedata.get_file_data(file)).hexdigest()
    return get_gamedata_hashes.gamedata_hashes


def get_savedata_hashes() -> {}:
    """ Gets a hash table for the unmodded contents of savedataformat.sarc """
    if not hasattr(get_savedata_hashes, 'savedata_hashes'):
        get_savedata_hashes.savedata_hashes = {}
        savedata = get_stock_savedata()
        for file in savedata.list_files():
            get_savedata_hashes.savedata_hashes[file] = xxhash.xxh32(
                savedata.get_file_data(file)).hexdigest()
    return get_savedata_hashes.savedata_hashes


def inject_gamedata_into_bootup(bgdata: sarc.SARCWriter, bootup_path: Path = None) -> int:
    """
    Packs a gamedata SARC into Bootup.pack and returns the RSTB size of the new gamedata.sarc

    :param bgdata: A SARCWriter for the new gamedata
    :type bgdata: class:`sarc.SARCWriter`
    :param bootup_path: Path to the Bootup.pack to update, defaults to a master BCML copy
    :type bootup_path: class:`pathlib.Path`, optional
    :returns: Returns the RSTB size of the new gamedata.sarc
    :rtype: int
    """
    if not bootup_path:
        master_boot = util.get_master_modpack_dir() / 'content' / 'Pack' / 'Bootup.pack'
        bootup_path = master_boot if master_boot.exists() \
            else util.get_game_file('Pack/Bootup.pack')
    with bootup_path.open('rb') as b_file:
        bootup_pack = sarc.read_file_and_make_sarc(b_file)
    new_pack = sarc.make_writer_from_sarc(bootup_pack)
    new_pack.delete_file('GameData/gamedata.ssarc')
    gamedata_bytes = bgdata.get_bytes()
    new_pack.add_file('GameData/gamedata.ssarc',
                      util.compress(gamedata_bytes))
    (util.get_master_modpack_dir() / 'content' /
     'Pack').mkdir(parents=True, exist_ok=True)
    with (util.get_master_modpack_dir() / 'content' / 'Pack' / 'Bootup.pack').open('wb') as b_file:
        new_pack.write(b_file)
    return rstb.SizeCalculator().calculate_file_size_with_ext(gamedata_bytes, True, '.sarc')


def inject_savedata_into_bootup(bgsvdata: sarc.SARCWriter, bootup_path: Path = None) -> int:
    """
    Packs a savedata SARC into Bootup.pack and returns the RSTB size of the new savedataformat.sarc

    :param bgsvdata: A SARCWriter for the new savedata
    :type bgsvdata: class:`sarc.SARCWriter`
    :param bootup_path: Path to the Bootup.pack to update, defaults to a master BCML copy
    :type bootup_path: class:`pathlib.Path`, optional
    :returns: Returns the RSTB size of the new savedataformat.sarc
    :rtype: int
    """
    if not bootup_path:
        master_boot = util.get_master_modpack_dir() / 'content' / 'Pack' / 'Bootup.pack'
        bootup_path = master_boot if master_boot.exists() \
            else util.get_game_file('Pack/Bootup.pack')
    with bootup_path.open('rb') as b_file:
        bootup_pack = sarc.read_file_and_make_sarc(b_file)
    new_pack = sarc.make_writer_from_sarc(bootup_pack)
    new_pack.delete_file('GameData/savedataformat.ssarc')
    savedata_bytes = bgsvdata.get_bytes()
    new_pack.add_file('GameData/savedataformat.ssarc',
                      util.compress(savedata_bytes))
    (util.get_master_modpack_dir() / 'content' / 'Pack').mkdir(parents=True, exist_ok=True)
    with (util.get_master_modpack_dir() / 'content' / 'Pack' / 'Bootup.pack').open('wb') as b_file:
        new_pack.write(b_file)
    return rstb.SizeCalculator().calculate_file_size_with_ext(savedata_bytes, True, '.sarc')


def is_savedata_modded(savedata: sarc.SARC) -> {}:
    """
    Detects if any .bgsvdata file has been modified.

    :param savedata: The saveformatdata.sarc to check for modification.
    :type savedata: class:`sarc.SARC`
    :returns: Returns True if any .bgsvdata in the given savedataformat.sarc has been modified.
    :rtype: bool
    """
    hashes = get_savedata_hashes()
    sv_files = sorted(savedata.list_files())
    fix_slash = '/' if not sv_files[0].startswith('/') else ''
    modded = False
    for svdata in sv_files[0:-2]:
        svdata_bytes = savedata.get_file_data(svdata).tobytes()
        svdata_hash = xxhash.xxh32(svdata_bytes).hexdigest()
        del svdata_bytes
        if not modded:
            modded = fix_slash + \
                svdata not in hashes or svdata_hash != hashes[fix_slash + svdata]
    return modded


def _bgdata_from_bytes(file: str, game_dict: dict) -> {}:
    return byml.Byml(game_dict[file]).parse()


def consolidate_gamedata(gamedata: sarc.SARC) -> {}:
    """
    Consolidates all game data in a game data SARC

    :return: Returns a dict of all game data entries in a SARC
    :rtype: dict of str: list
    """
    data = {}
    pool = Pool(processes=cpu_count())
    game_dict = {}
    for file in gamedata.list_files():
        game_dict[file] = gamedata.get_file_data(file).tobytes()
    results = pool.map(
        partial(_bgdata_from_bytes, game_dict=game_dict),
        gamedata.list_files()
    )
    pool.close()
    pool.join()
    del game_dict
    del gamedata
    for result in results:
        util.dict_merge(data, result)
    return data


def diff_gamedata_type(data_type: str, mod_data: dict, stock_data: dict) -> {}:
    """
    Logs the changes of a certain data type made to modded gamedata
    """
    stock_entries = [entry['DataName'] for entry in stock_data[data_type]]
    diffs = {}
    for entry in mod_data[data_type]:
        if entry['DataName'] not in stock_entries \
           or entry != stock_data[data_type][stock_entries.index(entry['DataName'])]:
            diffs[entry['DataName']] = deepcopy(entry)
    return {data_type: diffs}


def get_modded_gamedata_entries(gamedata: sarc.SARC) -> {}:
    """
    Gets all of the modified gamedata entries in a dict of modded gamedata contents.

    :param modded_bgdata: A dict of modified .bgdata files and their contents.
    :type modded_bgdata: dict
    :returns: Returns a dictionary with each data type and the modified entries for it.
    :rtype: dict of str: dict of str: dict
    """
    stock_data = consolidate_gamedata(get_stock_gamedata())
    mod_data = consolidate_gamedata(gamedata)
    diffs = {}
    pool = Pool(cpu_count())
    results = pool.map(
        partial(diff_gamedata_type, mod_data=mod_data, stock_data=stock_data),
        list(mod_data.keys())
    )
    pool.close()
    pool.join()
    for result in results:
        _, entries = list(result.items())[0]
        if entries:
            diffs.update(result)
    return diffs


def get_modded_savedata_entries(savedata: sarc.SARC) -> []:
    """
    Gets all of the modified savedata entries in a dict of modded savedata contents.

    :param savedata: The saveformatdata.sarc to search for modded entries.
    :type savedata: class:`sarc.SARC`
    :return: Returns a list of modified savedata entries.
    :rtype: list
    """
    ref_savedata = get_stock_savedata()
    ref_hashes = []
    new_entries = []
    for file in sorted(ref_savedata.list_files())[0:-2]:
        for item in byml.Byml(ref_savedata.get_file_data(file).tobytes()).parse()['file_list'][1]:
            ref_hashes.append(item['HashValue'])
    for file in sorted(savedata.list_files())[0:-2]:
        for item in byml.Byml(savedata.get_file_data(file).tobytes()).parse()['file_list'][1]:
            if item['HashValue'] not in ref_hashes:
                new_entries.append(item)
    return new_entries


def get_gamedata_mods() -> List[BcmlMod]:
    """ Gets a list of all installed mods that modify gamedata """
    gdata_mods = [mod for mod in util.get_installed_mods() if (
        mod.path / 'logs' / 'gamedata.yml').exists()]
    return sorted(gdata_mods, key=lambda mod: mod.priority)


def get_savedata_mods() -> List[BcmlMod]:
    """ Gets a list of all installed mods that modify save data """
    sdata_mods = [mod for mod in util.get_installed_mods() if (
        mod.path / 'logs' / 'savedata.yml').exists()]
    return sorted(sdata_mods, key=lambda mod: mod.priority)


def merge_gamedata(verbose: bool = False, force: bool = False):
    """ Merges installed gamedata mods and saves the new Bootup.pack, fixing the RSTB if needed """
    mods = get_gamedata_mods()
    glog_path = util.get_master_modpack_dir() / 'logs' / 'gamedata.log'
    if not mods:
        print('No gamedata merging necessary.')
        if glog_path.exists():
            glog_path.unlink()
        if (util.get_master_modpack_dir() / 'logs' / 'gamedata.sarc').exists():
            (util.get_master_modpack_dir() / 'logs' / 'gamedata.sarc').unlink()
        return
    if glog_path.exists() and not force:
        with glog_path.open('r') as l_file:
            if xxhash.xxh32(str(mods)).hexdigest() == l_file.read():
                print('No gamedata merging necessary.')
                return

    modded_entries = {}
    loader = yaml.CSafeLoader
    yaml_util.add_constructors(loader)
    print('Loading gamedata mods...')
    for mod in mods:
        with (mod.path / 'logs' / 'gamedata.yml').open('r') as g_file:
            yml = yaml.load(g_file, Loader=loader)
            for data_type in yml:
                if data_type not in modded_entries:
                    modded_entries[data_type] = {}
                modded_entries[data_type].update(yml[data_type])
                if verbose:
                    print(f'  Added entries for {data_type} from {mod.name}')

    gamedata = get_stock_gamedata()
    merged_entries = {}

    print('Loading stock gamedata...')
    for yml in gamedata.list_files():
        base_yml = byml.Byml(gamedata.get_file_data(yml).tobytes()).parse()
        for data_type in base_yml:
            if data_type not in merged_entries:
                merged_entries[data_type] = []
            merged_entries[data_type].extend(base_yml[data_type])

    print('Merging changes...')
    for data_type in merged_entries:
        if data_type in modded_entries:
            for entry in [entry for entry in merged_entries[data_type]
                          if entry['DataName'] in modded_entries[data_type]]:
                i = merged_entries[data_type].index(entry)
                if verbose:
                    print(f'  {entry["DataName"]} has been modified')
                merged_entries[data_type][i] = deepcopy(
                    modded_entries[data_type][entry['DataName']])
            print(f'Merged modified {data_type} entries')

    for data_type in modded_entries:
        for entry in [entry for entry in modded_entries[data_type]
                      if entry not in [entry['DataName'] for entry in merged_entries[data_type]]]:
            if verbose:
                print(f'  {entry} has been added')
            merged_entries[data_type].append(modded_entries[data_type][entry])
        print(f'Merged new {data_type} entries')

    print('Creating and injecting new gamedata.sarc...')
    new_gamedata = sarc.SARCWriter(True)
    for data_type in merged_entries:
        num_files = ceil(len(merged_entries[data_type]) / 4096)
        for i in range(num_files):
            end_pos = (i+1) * 4096
            if end_pos > len(merged_entries[data_type]):
                end_pos = len(merged_entries[data_type])
            buf = BytesIO()
            byml.Writer(
                {data_type: merged_entries[data_type][i*4096:end_pos]}, be=True).write(buf)
            new_gamedata.add_file(f'/{data_type}_{i}.bgdata', buf.getvalue())
    bootup_rstb = inject_gamedata_into_bootup(new_gamedata)
    (util.get_master_modpack_dir() / 'logs').mkdir(parents=True, exist_ok=True)
    with (util.get_master_modpack_dir() / 'logs' / 'gamedata.sarc').open('wb') as g_file:
        new_gamedata.write(g_file)

    print('Updating RSTB...')
    rstable.set_size('GameData/gamedata.sarc', bootup_rstb)

    glog_path.parent.mkdir(parents=True, exist_ok=True)
    with glog_path.open('w', encoding='utf-8') as l_file:
        l_file.write(xxhash.xxh32(str(mods)).hexdigest())


def merge_savedata(verbose: bool = False, force: bool = False):
    """ Merges install savedata mods and saves the new Bootup.pack, fixing the RSTB if needed"""
    mods = get_savedata_mods()
    slog_path = util.get_master_modpack_dir() / 'logs' / 'savedata.log'
    if not mods:
        print('No gamedata merging necessary.')
        if slog_path.exists():
            slog_path.unlink()
        if (util.get_master_modpack_dir() / 'logs' / 'savedata.sarc').exists():
            (util.get_master_modpack_dir() / 'logs' / 'savedata.sarc').unlink()
        return
    if slog_path.exists() and not force:
        with slog_path.open('r') as l_file:
            if xxhash.xxh32(str(mods)).hexdigest() == l_file.read():
                print('No savedata merging necessary.')
                return

    new_entries = []
    new_hashes = []
    loader = yaml.CSafeLoader
    yaml_util.add_constructors(loader)
    print('Loading savedata mods...')
    for mod in mods:
        with open(mod.path / 'logs' / 'savedata.yml') as s_file:
            yml = yaml.load(s_file, Loader=loader)
            for entry in yml:
                if entry['HashValue'] in new_hashes:
                    continue
                else:
                    new_entries.append(entry)
                    new_hashes.append(entry['HashValue'])
                    if verbose:
                        print(f'  Added {entry["DataName"]} from {mod.name}')

    savedata = get_stock_savedata()
    merged_entries = []
    save_files = sorted(savedata.list_files())[0:-2]

    print('Loading stock savedata...')
    for file in save_files:
        merged_entries.extend(byml.Byml(savedata.get_file_data(
            file).tobytes()).parse()['file_list'][1])

    print('Merging changes...')
    merged_entries.extend(new_entries)
    merged_entries.sort(key=lambda x: x['HashValue'])

    special_bgsv = [
        savedata.get_file_data('/saveformat_6.bgsvdata').tobytes(),
        savedata.get_file_data('/saveformat_7.bgsvdata').tobytes(),
    ]

    print('Creating and injecting new savedataformat.sarc...')
    new_savedata = sarc.SARCWriter(True)
    num_files = ceil(len(merged_entries) / 8192)
    for i in range(num_files):
        end_pos = (i+1) * 8192
        if end_pos > len(merged_entries):
            end_pos = len(merged_entries)
        buf = BytesIO()
        byml.Writer({
            'file_list': [
                {
                    'IsCommon': False,
                    'IsCommonAtSameAccount': False,
                    'IsSaveSecureCode': True,
                    'file_name': 'game_data.sav'
                },
                merged_entries[i*8192:end_pos]
            ],
            'save_info': [
                {
                    'directory_num': byml.Int(8),
                    'is_build_machine': True,
                    'revision': byml.Int(18203)
                }
            ]
        }, True).write(buf)
        new_savedata.add_file(f'/saveformat_{i}.bgsvdata', buf.getvalue())
    new_savedata.add_file(f'/saveformat_{num_files}.bgsvdata', special_bgsv[0])
    new_savedata.add_file(
        f'/saveformat_{num_files + 1}.bgsvdata', special_bgsv[1])
    bootup_rstb = inject_savedata_into_bootup(new_savedata)
    (util.get_master_modpack_dir() / 'logs').mkdir(parents=True, exist_ok=True)
    with (util.get_master_modpack_dir() / 'logs' / 'savedata.sarc').open('wb') as s_file:
        new_savedata.write(s_file)

    print('Updating RSTB...')
    rstable.set_size('GameData/savedataformat.sarc', bootup_rstb)

    slog_path.parent.mkdir(parents=True, exist_ok=True)
    with slog_path.open('w', encoding='utf-8') as l_file:
        l_file.write(xxhash.xxh32(str(mods)).hexdigest())


def get_stock_actorinfo() -> dict:
    """ Gets the unmodded contents of ActorInfo.product.sbyml """
    actorinfo = util.get_game_file('Actor/ActorInfo.product.sbyml')
    with actorinfo.open('rb') as a_file:
        return byml.Byml(util.decompress(a_file.read())).parse()


def get_modded_actors(actorinfo: dict) -> dict:
    """
    Gets new or changed actors in modded ActorInfo data and their hashes

    :param actorinfo: A dict representing the contents of ActorInfo.product.sbyml
    :type actorinfo: dict
    :return: Returns a dict of actor hashes and their associated actor info
    :rtype: dict of int: dict
    """
    stock_actors = get_stock_actorinfo()
    modded_actors = {}
    for actor in actorinfo['Actors']:
        actor_hash = zlib.crc32(actor['name'].encode())
        if actor_hash not in stock_actors['Hashes']:
            modded_actors[actor_hash] = actor
        elif actor != stock_actors['Actors'][stock_actors['Hashes'].index(actor_hash)]:
            stock_actor = stock_actors['Actors'][stock_actors['Hashes'].index(
                actor_hash)]
            modded_actors[actor_hash] = {}
            for key, value in actor.items():
                if key not in stock_actor or value != stock_actor[key]:
                    modded_actors[actor_hash][key] = value
    return modded_actors


def get_actorinfo_mods() -> List[BcmlMod]:
    """ Gets a list of all installed mods that modify ActorInfo.product.sbyml """
    actor_mods = [mod for mod in util.get_installed_mods()\
                  if (mod.path / 'logs' / 'actorinfo.yml').exists()]
    return sorted(actor_mods, key=lambda mod: mod.priority)


def merge_actorinfo(verbose: bool = False):
    """Merges installed changes to actor info"""
    mods = get_actorinfo_mods()
    actor_path = (util.get_master_modpack_dir() / 'content' /
                  'Actor' / 'ActorInfo.product.sbyml')
    if not mods:
        print('No actor info merging necessary.')
        if actor_path.exists():
            actor_path.unlink()
        return

    print('Loading modded actor info...')
    modded_actors = {}
    loader = yaml.CSafeLoader
    yaml_util.add_constructors(loader)
    for mod in mods:
        with (mod.path / 'logs' / 'actorinfo.yml').open('r', encoding='utf-8') as a_file:
            entries = yaml.load(a_file, Loader=loader)
            util.dict_merge(modded_actors, entries, overwrite_lists=True)
            if verbose:
                print(f'Loaded {len(entries)} entries from {mod.name}')
            del entries
    print('Loading unmodded actor info...')
    actorinfo = get_stock_actorinfo()

    print('Merging changes...')
    for actor_hash, actor_info in modded_actors.items():
        if actor_hash in actorinfo['Hashes']:
            idx = actorinfo['Hashes'].index(actor_hash)
            util.dict_merge(actorinfo['Actors'][idx],
                            actor_info, overwrite_lists=True)
            if verbose:
                print(f'  Updated entry for {actorinfo["Actors"][idx]}')
        else:
            actorinfo['Hashes'].append(actor_hash)
            actorinfo['Actors'].append(actor_info)
            if verbose:
                print(f'  Added entry for {actor_info["name"]}')

    print('Sorting new actor info...')
    actorinfo['Hashes'].sort()
    actorinfo['Hashes'] = list(map(lambda x: byml.Int(
        x) if x < 2147483648 else byml.UInt(x), actorinfo['Hashes']))
    actorinfo['Actors'].sort(
        key=lambda x: zlib.crc32(x['name'].encode('utf-8')))

    print('Saving new actor info...')
    buf = BytesIO()
    byml.Writer(actorinfo, True).write(buf)
    actor_path.parent.mkdir(parents=True, exist_ok=True)
    actor_path.write_bytes(util.compress(buf.getvalue()))
    print('Actor info merged successfully')


class GameDataMerger(mergers.Merger):
    NAME: str = 'gamedata'

    def __init__(self):
        super().__init__(
            'game data merge',
            'Merges changes to gamedata.sarc',
            'gamedata.yml', options={}
        )

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[Path, str]]):
        if 'content/Pack/Bootup.pack//GameData/gamedata.ssarc' in modded_files:
            with (mod_dir / 'content' / 'Pack' / 'Bootup.pack').open('rb') as bootup_file:
                bootup_sarc = sarc.read_file_and_make_sarc(bootup_file)
            return get_modded_gamedata_entries(
                sarc.SARC(
                    util.decompress(
                        bootup_sarc.get_file_data('GameData/gamedata.ssarc').tobytes()
                    )
                )
            )
        else:
            return {}

    def log_diff(self, mod_dir: Path, diff_material: Union[dict, List[Path]]):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(mod_dir, diff_material)
        if diff_material:
            with (mod_dir / 'logs' / self._log_name).open('w', encoding='utf-8') as log:
                dumper = yaml.CSafeDumper
                yaml_util.add_representers(dumper)
                yaml.dump(diff_material, log, Dumper=dumper, allow_unicode=True, encoding='utf-8',
                          default_flow_style=None)

    def get_mod_diff(self, mod: BcmlMod):
        if self.is_mod_logged(mod):
            with (mod.path / 'logs' / self._log_name).open('r', encoding='utf-8') as log:
                loader = yaml.CSafeLoader
                yaml_util.add_constructors(loader)
                return yaml.load(log, Loader=loader)
        else:
            return {}

    def get_all_diffs(self):
        diffs = []
        for mod in get_gamedata_mods():
            diffs.append(self.get_mod_diff(mod))
        return diffs

    def consolidate_diffs(self, diffs: list):
        all_diffs = {}
        for diff in diffs:
            util.dict_merge(all_diffs, diff, overwrite_lists=True)
        return all_diffs

    def perform_merge(self):
        force = False
        if 'force' in self._options:
            force = self._options['force']
        merge_gamedata(force=force)

    def get_checkbox_options(self):
        return [('force', 'Remerge game data even if no changes detected')]

    @staticmethod
    def is_bootup_injector():
        return True

    def get_bootup_injection(self):
        tmp_sarc = util.get_master_modpack_dir() / 'logs' / 'gamedata.sarc'
        if tmp_sarc.exists():
            return (
                'GameData/gamedata.ssarc',
                util.compress(tmp_sarc.read_bytes())
            )
        else:
            return


class SaveDataMerger(mergers.Merger):
    NAME: str = 'savedata'

    def __init__(self):
        super().__init__('save data merge', 'Merge changes to savedataformat.ssarc', 'savedata.yml')

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[Path, str]]):
        if 'content/Pack/Bootup.pack//GameData/savedataformat.ssarc' in modded_files:
            with (mod_dir / 'content' / 'Pack' / 'Bootup.pack').open('rb') as bootup_file:
                bootup_sarc = sarc.read_file_and_make_sarc(bootup_file)
            return get_modded_savedata_entries(
                sarc.SARC(
                    util.decompress(
                        bootup_sarc.get_file_data('GameData/savedataformat.ssarc').tobytes()
                    )
                )
            )
        else:
            return []

    def log_diff(self, mod_dir: Path, diff_material):
        if isinstance(diff_material[0], Path):
            diff_material = self.generate_diff(mod_dir, diff_material)
        if diff_material:
            with (mod_dir / 'logs' / self._log_name).open('w', encoding='utf-8') as log:
                dumper = yaml.CSafeDumper
                yaml_util.add_representers(dumper)
                yaml.dump(diff_material, log, Dumper=dumper, allow_unicode=True, encoding='utf-8',
                          default_flow_style=None)

    def get_mod_diff(self, mod: BcmlMod):
        if self.is_mod_logged(mod):
            with (mod.path / 'logs' / self._log_name).open('r', encoding='utf-8') as log:
                loader = yaml.CSafeLoader
                yaml_util.add_constructors(loader)
                return yaml.load(log, Loader=loader)
        else:
            return []

    def get_all_diffs(self):
        diffs = []
        for mod in get_savedata_mods():
            diffs.append(self.get_mod_diff(mod))
        return diffs

    def consolidate_diffs(self, diffs: list):
        all_diffs = []
        hashes = []
        for diff in reversed(diffs):
            for entry in diff:
                if entry['HashValue'] not in hashes:
                    all_diffs.append(entry)
        return all_diffs

    def perform_merge(self):
        force = False
        if 'force' in self._options:
            force = self._options['force']
        merge_savedata(force=force)

    def get_checkbox_options(self):
        return [('force', 'Remerge save data even if no changes detected')]

    @staticmethod
    def is_bootup_injector():
        return True

    def get_bootup_injection(self):
        tmp_sarc = util.get_master_modpack_dir() / 'logs' / 'savedata.sarc'
        if tmp_sarc.exists():
            return (
                'GameData/savedataformat.ssarc',
                util.compress(tmp_sarc.read_bytes())
            )
        else:
            return


class ActorInfoMerger(mergers.Merger):
    NAME: str = 'actors'

    def __init__(self):
        super().__init__('actor info merge', 'Merges changes to ActorInfo.product.byml',
                         'actorinfo.yml ', {})

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[Path, str]]):
        try:
            actor_file = next(iter([file for file in modded_files \
                               if Path(file).name == 'ActorInfo.product.sbyml']))
        except StopIteration:
            return {}
        actor_info = byml.Byml(util.decompress_file(str(actor_file))).parse()
        print('Detecting modified actor info entries...')
        return get_modded_actors(actor_info)

    def log_diff(self, mod_dir: Path, diff_material: Union[dict, list]):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(Path, diff_material)
        if diff_material:
            with (mod_dir / 'logs' / self._log_name).open('w', encoding='utf-8') as log:
                dumper = yaml.CSafeDumper
                yaml_util.add_representers(dumper)
                yaml.dump(diff_material, log, Dumper=dumper, allow_unicode=True, encoding='utf-8',
                          default_flow_style=None)

    def get_mod_diff(self, mod: BcmlMod):
        if self.is_mod_logged(mod):
            with (mod.path / 'logs' / self._log_name).open('r', encoding='utf-8') as log:
                loader = yaml.CSafeLoader
                yaml_util.add_constructors(loader)
                return yaml.load(log, Loader=loader)
        else:
            return {}

    def get_all_diffs(self):
        diffs = []
        for mod in get_gamedata_mods():
            diffs.append(self.get_mod_diff(mod))
        return diffs

    def consolidate_diffs(self, diffs: list):
        all_diffs = {}
        for diff in diffs:
            all_diffs.update(diff)
        return all_diffs

    def perform_merge(self):
        merge_actorinfo()

    def get_checkbox_options(self):
        return []



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\macad\Documents\Git\BCML-2\.vscode\options.ui',
# licensing of 'c:\Users\macad\Documents\Git\BCML-2\.vscode\options.ui' applies.
#
# Created: Thu Sep 12 14:31:57 2019
#      by: pyside2-uic  running on PySide2 5.13.0
#
# WARNING! All changes made in this file will be lost!

from bcml import mergers
from PySide2 import QtCore, QtGui, QtWidgets

class Ui_OptionsDialog(object):
    def setupUi(self, OptionsDialog):
        OptionsDialog.setObjectName("OptionsDialog")
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(OptionsDialog.sizePolicy().hasHeightForWidth())
        OptionsDialog.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(OptionsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(OptionsDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.checkboxes = []
        for merger_class in mergers.get_mergers():
            merger = merger_class()
            chkMerger = QtWidgets.QCheckBox(OptionsDialog)
            chkMerger.setObjectName('chkDisable' + merger.NAME)
            chkMerger.setText('Disable ' + merger.friendly_name())
            setattr(chkMerger, 'disable_name', merger.NAME)
            self.verticalLayout.addWidget(chkMerger)
            self.checkboxes.append(chkMerger)
            for option in merger.get_checkbox_options():
                chkOption = QtWidgets.QCheckBox(OptionsDialog)
                chkOption.setObjectName('chk' + merger.NAME + option[0])
                setattr(chkOption, 'option_name', option[0])
                setattr(chkOption, 'merger', merger.NAME)
                chkOption.setText(option[1])
                self.verticalLayout.addWidget(chkOption)
                self.checkboxes.append(chkOption)
        self.buttonBox = QtWidgets.QDialogButtonBox(OptionsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(OptionsDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), OptionsDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), OptionsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(OptionsDialog)

    def retranslateUi(self, OptionsDialog):
        OptionsDialog.setWindowTitle(QtWidgets.QApplication.translate("OptionsDialog", "Advanced Options", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("OptionsDialog", "Select Advanced Mod Options", None, -1))




# pylint: skip-file

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(276, 277)
        self.verticalLayout = QtWidgets.QVBoxLayout(AboutDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lblLogo = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblLogo.sizePolicy().hasHeightForWidth())
        self.lblLogo.setSizePolicy(sizePolicy)
        self.lblLogo.setMinimumSize(QtCore.QSize(0, 108))
        self.lblLogo.setObjectName("lblLogo")
        self.verticalLayout.addWidget(self.lblLogo)
        self.label_2 = QtWidgets.QLabel(AboutDialog)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.lblVersion = QtWidgets.QLabel(AboutDialog)
        self.lblVersion.setObjectName("lblVersion")
        self.verticalLayout.addWidget(self.lblVersion)
        self.lblSites = QtWidgets.QLabel(AboutDialog)
        self.lblSites.setTextFormat(QtCore.Qt.RichText)
        self.lblSites.setObjectName("lblSites")
        self.verticalLayout.addWidget(self.lblSites)
        self.label = QtWidgets.QLabel(AboutDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.label_3 = QtWidgets.QLabel(AboutDialog)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(AboutDialog)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btnClose = QtWidgets.QPushButton(AboutDialog)
        self.btnClose.setObjectName("btnClose")
        self.horizontalLayout.addWidget(self.btnClose)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(AboutDialog)
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(QtWidgets.QApplication.translate("AboutDialog", "About BCML", None, -1))
        self.lblLogo.setText(QtWidgets.QApplication.translate("AboutDialog", "TextLabel", None, -1))
        self.label_2.setText(QtWidgets.QApplication.translate("AboutDialog", "A mod installer and manager for BotW mods on Cemu", None, -1))
        self.lblVersion.setText(QtWidgets.QApplication.translate("AboutDialog", "{version}", None, -1))
        self.lblSites.setText(QtWidgets.QApplication.translate("AboutDialog", "<a href=\"https://gamebanana.com/tools/6624\">View on GameBanana</a> <a href=\"https://github.com/NiceneNerd/BCML\">View on GitHub</a>", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("AboutDialog", "BCML © 2019 Caleb Smith, License: GPLv3+", None, -1))
        self.label_3.setText(QtWidgets.QApplication.translate("AboutDialog", "7z.exe © 2019 Ignor Pavolv, License: LGPLv3+", None, -1))
        self.label_4.setText(QtWidgets.QApplication.translate("AboutDialog", "msyt.exe © 2018 Kyle Clemens, License: MIT", None, -1))
        self.btnClose.setText(QtWidgets.QApplication.translate("AboutDialog", "Close", None, -1))

